from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Sequence, TypeVar

import numpy as np

from whitecanvas import theme
from whitecanvas.canvas.dataframe import _misc
from whitecanvas.canvas.dataframe._base import AggMethods, BaseCatPlotter, CatIterator
from whitecanvas.layers import tabular as _lt
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.types import (
    ColorType,
    Hatch,
    HistBinType,
    Orientation,
    Symbol,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas._base import CanvasBase
    from whitecanvas.canvas.dataframe._stacked_cat import StackedCatPlotter
    from whitecanvas.layers.tabular._box_like import _BoxLikeMixin
    from whitecanvas.layers.tabular._dataframe import DataFrameWrapper

    NStr = str | Sequence[str]

_C = TypeVar("_C", bound="CanvasBase")
_DF = TypeVar("_DF")


class _Aggregator(Generic[_C, _DF]):
    def __init__(self, method: str, plotter: OneAxisCatPlotter[_C, _DF] = None):
        self._method = method
        self._plotter = plotter

    def __get__(self, ins: _C, owner) -> Self:
        return _Aggregator(self._method, ins)

    def __repr__(self) -> str:
        return f"Aggregator<{self._method}>"

    def __call__(self) -> OneAxisCatAggPlotter[_C, _DF]:
        """
        Aggregate the values before plotting it.

        Examples
        --------
        Calculate mean of "value" for each "group" and connect them with lines.
        >>> canvas.cat_x(df, "group", "value").mean().add_line()
        """
        plotter = self._plotter
        if plotter is None:
            raise TypeError("Cannot call this method from a class.")
        if self._method == "size":
            value = "size"
        elif plotter._value is None:
            raise ValueError("Value column is not specified.")
        else:
            value = plotter._value
        return OneAxisCatAggPlotter(
            plotter._canvas(),
            plotter._cat_iter,
            offset=plotter._offset,
            value=value,
            method=self._method,
            orient=plotter._orient,
        )


class _GroupAggregator(Generic[_C, _DF]):
    def __init__(self, method: str, plotter: OneAxisCatPlotter[_C, _DF] = None):
        self._method = method
        self._plotter = plotter

    def __get__(self, ins: _C, owner) -> Self:
        return _GroupAggregator(self._method, ins)

    def __repr__(self) -> str:
        return f"GroupAggregator<{self._method}>"

    def __call__(self, by: str | tuple[str, ...]) -> OneAxisCatPlotter[_C, _DF]:
        """Aggregate the values for each group before plotting it."""
        plotter = self._plotter
        if isinstance(by, str):
            by = (by,)
        elif len(by) == 0:
            raise ValueError("No column is specified for grouping.")
        return type(plotter)(
            plotter._canvas(),
            plotter._df.agg_by((*plotter._offset, *by), [plotter._value], self._method),
            offset=plotter._offset,
            value=plotter._value,
            update_labels=plotter._update_labels,
        )


class OneAxisCatPlotter(BaseCatPlotter[_C, _DF]):
    _orient: Orientation

    def __init__(
        self,
        canvas: _C,
        df: _DF,
        offset: str | tuple[str, ...] | None,
        value: str | None,
        update_labels: bool = False,
        numeric: bool = False,
        sort_func: Callable[[tuple], int] | None = None,
    ):
        super().__init__(canvas, df)
        if isinstance(offset, str):
            offset = (offset,)
        elif offset is None:
            offset = ()
        else:
            if any(not isinstance(o, str) for o in offset):
                raise TypeError(
                    "Category column(s) must be specified by a string or a sequence "
                    f"of strings, got {offset!r}."
                )
        if numeric and offset is not None:
            if len(offset) != 1:
                raise ValueError(
                    "Numerical axis with multiple categories is not supported."
                )
            arr = self._df[offset[0]]
            if arr.dtype.kind not in "iuf":
                raise ValueError(
                    f"Column {offset[0]!r} cannot be interpreted as a numerical axis "
                    f"(dtype={arr.dtype!r})."
                )
        # check dtype?
        # for col in offset:
        #     arr = self._df[col]
        #     if arr.dtype.kind in "fc":
        #         raise ValueError(
        #             f"Column {col!r} cannot be interpreted as a categorical column "
        #             f"(dtype={arr.dtype!r})."
        #         )
        self._offset: tuple[str, ...] = offset
        self._cat_iter = CatIterator(
            self._df,
            offsets=offset,
            numeric=numeric,
            sort_func=sort_func,
        )
        self._value = value
        self._update_labels = update_labels
        if update_labels:
            if value is not None:
                self._update_axis_labels(value)
            if not numeric:
                pos, label = self._cat_iter.axis_ticks()
                if self._orient.is_vertical:
                    canvas.x.ticks.set_labels(pos, label)
                    canvas.x.lim = (np.min(pos) - 0.5, np.max(pos) + 0.5)
                else:
                    canvas.y.ticks.set_labels(pos, label)
                    canvas.y.lim = (np.min(pos) - 0.5, np.max(pos) + 0.5)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(offset={self._offset!r}, value={self._value!r}, "
            f"orient={self._orient!r})"
        )

    def _update_axis_labels(self, value_label: str) -> None:
        """Update the x and y labels using the column names"""
        canvas = self._canvas()
        offset_label = self._cat_iter.axis_label()
        if self._orient.is_vertical:
            canvas.x.label.text = offset_label
            canvas.y.label.text = value_label
        else:
            canvas.x.label.text = value_label
            canvas.y.label.text = offset_label

    def _get_value(self) -> str:
        if self._value is None:
            raise ValueError("Value column is not specified.")
        return self._value

    def stack(
        self,
        by: str | Sequence[str] | None = None,
    ) -> StackedCatPlotter[_C, _DF]:
        """
        Return a stacked plotter for data stacked by the given column(s).

        >>> canvas.cat_x(df, "time", "value").stack("region").add_bars()
        >>> canvas.cat_x(df, "time", "value").stack("region").add_bars(color="region")
        >>> canvas.cat_x(df, "time", "value").stack("region").add_bars(hatch="region")

        Parameters
        ----------
        by : str or sequence of str
            Column name(s) for stacking the data.
        """
        from whitecanvas.canvas.dataframe._stacked_cat import StackedCatPlotter

        return StackedCatPlotter(
            self._canvas(),
            df=self._df,
            offset=self._offset,
            value=self._value,
            orient=self._orient,
            stackby=by,
            update_labels=self._update_labels,
            numeric=self._cat_iter._numeric,
        )

    def melt(
        self,
        id_vars: str | Sequence[str] | None = None,
        value_vars: str | Sequence[str] | None = None,
        var_name: str | None = None,
        value_name: str | None = None,
    ) -> Self:
        """
        Melt the dataframe to long format and update the x/y column names.

        This method should be used without setting x or y.

        >>> canvas.cat_x(df).melt().add_stripplot()

        Parameters
        ----------
        id_vars : str or sequence of str
            Column name(s) to use as identifier variables.
        value_vars : str or sequence of str
            Column name(s) to use as value variables.
        var_name : str, optional
            Name of the variable column. If None, it will be set to "variable".
        value_name : str, optional
            Name of the value column. If None, it will be set to "value".
        """
        if self._value is not None:
            raise ValueError("Cannot melt the dataframe with x or y set.")
        if id_vars is None:
            id_vars = []
        if value_vars is None:
            value_vars = self._df.columns
        if var_name is None:
            var_name = "variable"
        if value_name is None:
            value_name = "value"
        df = self._df.melt(id_vars, value_vars, var_name, value_name)
        out = type(self)(self._canvas(), df, var_name, value_name, self._update_labels)
        out._df = df
        return out

    def sort_in_order(self, order: Sequence[Any]) -> Self:
        """
        Sort the categories in the given order.

        Parameters
        ----------
        order : sequence of any
            The order to sort the data.
        """
        order_normed = tuple(u if isinstance(u, tuple) else (u,) for u in order)
        return type(self)(
            self._canvas(), self._df, self._offset, self._value, self._update_labels,
            sort_func=_misc.make_sorter_manual(order_normed)
        )  # fmt: skip

    def sort(self, *, ascending: bool = True) -> Self:
        """
        Sort the categories in ascending or descending order.

        Parameters
        ----------
        ascending : bool, default True
            Whether to sort the data in ascending order.
        """
        return type(self)(
            self._canvas(),
            self._df,
            self._offset,
            self._value,
            self._update_labels,
            sort_func=_misc.sorter_ascending if ascending else _misc.sorter_descending,
        )

    ### 1-D categorical ###

    def add_violinplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | None = None,
        hatch: NStr | None = None,
        dodge: NStr | bool = True,
        extent: float = 0.8,
        shape: str = "both",
    ) -> _lt.DFViolinPlot[_DF]:
        """
        Add a categorical violin plot.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_violinplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_violinplot(
        ...     color="region", dodge=True,
        ... )

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float, optional
            Opacity of the layer.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        extent : float, default 0.8
            Width of the violins. Usually in range (0, 1].
        shape : str, default "both"
            Shape of the violins. Can be "both", "left", or "right".

        Returns
        -------
        DFViolinPlot
            Violin plot layer.
        """
        canvas = self._canvas()
        layer = _lt.DFViolinPlot(
            self._cat_iter, self._get_value(), name=name, color=color, hatch=hatch,
            dodge=dodge, extent=extent, shape=shape, orient=self._orient,
            backend=canvas._get_backend(),
        )  # fmt: skip
        self._post_add_boxlike(layer, color, alpha)
        return canvas.add_layer(layer)

    def add_boxplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | None = None,
        hatch: NStr | None = None,
        dodge: NStr | bool = True,
        capsize: float = 0.1,
        extent: float = 0.8,
    ) -> _lt.DFBoxPlot[_DF]:
        """
        Add a categorical box plot.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_boxplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_boxplot(
        ...     color="region", dodge=True,
        ... )

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float, optional
            Opacity of the layer.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        capsize : float, default 0.1
            Length of the caps as a fraction of the width of the box.
        extent : float, default 0.8
            Width of the violins. Usually in range (0, 1].

        Returns
        -------
        DFBoxPlot
            Box plot layer.
        """
        canvas = self._canvas()
        layer = _lt.DFBoxPlot(
            self._cat_iter, self._get_value(), name=name, color=color, hatch=hatch,
            dodge=dodge, orient=self._orient, capsize=capsize, extent=extent,
            backend=canvas._get_backend(),
        )  # fmt: skip
        self._post_add_boxlike(layer, color, alpha)
        return canvas.add_layer(layer)

    def add_pointplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | None = None,
        hatch: NStr | None = None,
        dodge: NStr | bool = True,
        capsize: float = 0.1,
    ) -> _lt.DFPointPlot[_DF]:
        """
        Add a categorical point plot (markers with error bars).

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_pointplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_pointplot(
        ...     color="region", dodge=True,
        ... )

        The default estimator and errors are mean and standard deviation. To change
        them, use `est_by_*` and `err_by_*` methods.

        >>> ### Use standard error x 2 (~95%) as error bars.
        >>> canvas.cat_x(df, "species", "weight").add_pointplot().err_by_se(scale=2.0)

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float, optional
            Opacity of the layer.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        capsize : float, default 0.1
            Length of the caps as a fraction of the width of the box.

        Returns
        -------
        DFPointPlot
            Point plot layer.
        """
        canvas = self._canvas()
        layer = _lt.DFPointPlot(
            self._cat_iter, self._get_value(), name=name, color=color, hatch=hatch,
            dodge=dodge, orient=self._orient, capsize=capsize,
            backend=canvas._get_backend(),
        )  # fmt: skip
        self._post_add_boxlike(layer, color, alpha)
        return canvas.add_layer(layer)

    def add_barplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | None = None,
        hatch: NStr | None = None,
        dodge: NStr | bool = True,
        capsize: float = 0.1,
        extent: float = 0.8,
    ) -> _lt.DFBarPlot[_DF]:
        """
        Add a categorical bar plot (bars with error bars).

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_barplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_barplot(
        ...     color="region", dodge=True
        ... )

        The default estimator and errors are mean and standard deviation. To change
        them, use `est_by_*` and `err_by_*` methods.

        >>> ### Use standard error x 2 (~95%) as error bars.
        >>> canvas.cat_x(df, "species", "weight").add_barplot().err_by_se(scale=2.0)

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float, optional
            Opacity of the layer.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        capsize : float, default 0.1
            Length of the caps as a fraction of the width of the box.
        extent : float, default 0.8
            Width of the violins. Usually in range (0, 1].

        Returns
        -------
        DFBarPlot
            Bar plot layer.
        """
        canvas = self._canvas()
        layer = _lt.DFBarPlot(
            self._cat_iter, self._get_value(), name=name, color=color, hatch=hatch,
            dodge=dodge, orient=self._orient, capsize=capsize, extent=extent,
            backend=canvas._get_backend(),
        )  # fmt: skip
        self._post_add_boxlike(layer, color, alpha)
        return canvas.add_layer(layer)

    def _post_add_boxlike(self, layer: _BoxLikeMixin, color, alpha):
        canvas = self._canvas()
        if color is not None and not layer._color_by.is_const():
            layer.update_color_palette(canvas._color_palette)
        elif color is None:
            layer.update_const(color=canvas._color_palette.next())
        if alpha is not None:
            layer.update_const(alpha=alpha)

    def add_stripplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | str | None = None,
        hatch: NStr | None = None,
        symbol: NStr | None = None,
        size: str | None = None,
        dodge: NStr | bool = False,
        extent: float = 0.5,
        seed: int | None = 0,
    ) -> _lt.DFMarkerGroups[_DF]:
        """
        Add a categorical strip plot.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_stripplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_stripplot(
        ...     color="region", dodge=True
        ... )

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float or str, optional
            Opacity of the markers. If str, it must be a numerical column name.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        extent : float, default 0.5
            Width of the violins. Usually in range (0, 1].
        seed : int, optional
            Random seed for jittering.

        Returns
        -------
        DFMarkerGroups
            Marker collection layer.
        """
        canvas = self._canvas()
        symbol = theme._default("markers.symbol", symbol)
        size = theme._default("markers.size", size)

        df = self._df
        splitby, dodge = _shared.norm_dodge_markers(
            df, self._offset, color, hatch, dodge
        )  # fmt: skip
        _map = self._cat_iter.prep_position_map(splitby, dodge)
        _extent = self._cat_iter.zoom_factor(dodge) * extent
        xj = _jitter.UniformJitter(splitby, _map, extent=_extent, seed=seed)
        yj = _jitter.IdentityJitter(self._get_value()).check(df)
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFMarkerGroups(
            df, xj, yj, name=name, color=color, hatch=hatch, orient=self._orient,
            symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        layer._init_color_for_canvas(color, alpha, canvas)
        return canvas.add_layer(layer)

    def add_markers(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | str | None = None,
        hatch: NStr | None = None,
        symbol: NStr | None = None,
        size: str | None = None,
        dodge: NStr | bool = False,
    ) -> _lt.DFMarkerGroups[_DF]:
        """Alias of `add_stripplot` with no jittering."""
        return self.add_stripplot(
            color=color, alpha=alpha, hatch=hatch, symbol=symbol, size=size,
            dodge=dodge, extent=0, seed=0, name=name,
        )  # fmt: skip

    def add_swarmplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | str | None = None,
        hatch: NStr | None = None,
        symbol: NStr | None = None,
        size: str | None = None,
        dodge: NStr | bool = False,
        extent: float = 0.8,
        sort: bool = False,
    ) -> _lt.DFMarkerGroups[_DF]:
        """
        Add a categorical swarm plot.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_swarmplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_swarmplot(
        ...     color="region", dodge=True
        ... )

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float or str, optional
            Opacity of the markers. If str, it must be a numerical column name.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        extent : float, default 0.8
            Width of the violins. Usually in range (0, 1].
        sort : bool, default False
            Whether to sort the data by value.

        Returns
        -------
        DFMarkerGroups
            Marker collection layer.
        """
        canvas = self._canvas()
        symbol = theme._default("markers.symbol", symbol)
        size = theme._default("markers.size", size)
        df = self._df
        splitby, dodge = _shared.norm_dodge_markers(
            df, self._offset, color, hatch, dodge
        )  # fmt: skip
        _map = self._cat_iter.prep_position_map(splitby, dodge)
        _extent = self._cat_iter.zoom_factor(dodge) * extent

        val = self._get_value()
        if sort:
            df = df.sort(val)
        lims = df[val].min(), df[val].max()
        xj = _jitter.SwarmJitter(splitby, _map, val, limits=lims, extent=_extent)
        yj = _jitter.IdentityJitter(val).check(df)
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFMarkerGroups(
            df, xj, yj, name=name, color=color, hatch=hatch, orient=self._orient,
            symbol=symbol, size=size, backend=canvas._get_backend(),
        )  # fmt: skip
        layer._init_color_for_canvas(color, alpha, canvas)
        return canvas.add_layer(layer)

    def add_rugplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | str | None = None,
        width: float | None = None,
        style: NStr | None = None,
        dodge: NStr | bool = True,
        extent: float = 0.8,
    ) -> _lt.DFRugGroups[_DF]:
        """
        Add a categorical rug plot.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_rugplot()

        >>> ### Color by column "region" with dodging.
        >>> canvas.cat_x(df, x="species", y="weight").add_rugplot(
        ...     color="region", dodge=True
        ... )

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        alpha : float or str, optional
            Opacity of the markers. If str, it must be a numerical column name.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        extent : float, default 0.8
            Width of the violins. Usually in range (0, 1].
        sort : bool, default False
            Whether to sort the data by value.

        Returns
        -------
        DFMarkerGroups
            Marker collection layer.
        """
        canvas = self._canvas()
        width = theme._default("line.width", width)

        df = self._df
        splitby, dodge = _shared.norm_dodge(
            df, self._offset, color, style, dodge=dodge
        )  # fmt: skip
        _map = self._cat_iter.prep_position_map(splitby, dodge)
        _extent = self._cat_iter.zoom_factor(dodge) * extent
        jitter = _jitter.CategoricalJitter(splitby, _map)
        layer = _lt.DFRugGroups.from_table(
            df, jitter, self._get_value(), name=name, color=color, orient=self._orient,
            extent=_extent, width=width, style=style, backend=canvas._get_backend(),
        )  # fmt: skip
        layer._init_color_for_canvas(color, alpha, canvas)
        return canvas.add_layer(layer)

    def add_heatmap_hist(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        dodge: NStr | bool = False,
        bins: HistBinType = "auto",
        range: tuple[float, float] | None = None,
    ) -> _lt.DFMultiHeatmap[_DF]:
        """
        Add a 1-D histogram heatmap.

        >>> ### Use "species" column as categories and "weight" column as values.
        >>> canvas.cat_x(df, x="species", y="weight").add_heatmap_hist()

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the heatmaps. Must be categorical.
        dodge : str, sequence of str or bool, optional
            Column name(s) for dodging. If True, column name(s) for splitting the data
            will all be used for dodging to avoid overlap.
        bins : int or sequence of int or str, default "auto"
            Number of bins.
        range : (float, float), optional
            The lower and upper range of the bins.

        Returns
        -------
        DFMultiHeatmap
            Heatmap layer.
        """
        canvas = self._canvas()
        layer = _lt.DFMultiHeatmap.build_hist_1d(
            self._cat_iter, self._offset, self._get_value(), color=color, bins=bins,
            name=name, range=range, palette=canvas._color_palette, orient=self._orient,
            dodge=dodge, backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(layer)

    # aggregators and group aggregators
    mean = _Aggregator("mean")
    median = _Aggregator("median")
    min = _Aggregator("min")
    max = _Aggregator("max")
    std = _Aggregator("std")
    sum = _Aggregator("sum")
    count = _Aggregator("size")
    first = _Aggregator("first")

    mean_for_each = _GroupAggregator("mean")
    median_for_each = _GroupAggregator("median")
    min_for_each = _GroupAggregator("min")
    max_for_each = _GroupAggregator("max")
    std_for_each = _GroupAggregator("std")
    sum_for_each = _GroupAggregator("sum")
    first_for_each = _GroupAggregator("first")


class OneAxisCatAggPlotter(BaseCatPlotter[_C, _DF]):
    """Class for plotting aggregated values of a single categorical axis."""

    def __init__(
        self,
        canvas: _C,
        cat_iter: CatIterator[_DF],
        offset: str | tuple[str, ...],
        value: str,
        method: AggMethods,
        orient: Orientation,
    ):
        super().__init__(canvas, cat_iter._df)
        self._offset = offset
        self._value = value
        self._agg_method = method
        self._orient = orient
        self._cat_iter = cat_iter

    def add_line(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        alpha: float | None = None,
        width: float | None = None,
        style: NStr | None = None,
        dodge: NStr | bool = False,
    ) -> _lt.DFLines[_DF]:
        """
        Add line that connect the aggregated values.

        >>> canvas.cat_x(df, "time", "value").mean().add_line()

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        width : float, optional
            Line width.
        style : str or sequence of str, optional
            Column name(s) for styling the lines. Must be categorical.

        Returns
        -------
        DFLines
            Line collection layer.
        """
        # TODO: support width: str
        canvas = self._canvas()
        df = self._df
        width = theme._default("line.width", width)

        _splitby, _dodge = _shared.norm_dodge(df, self._offset, color, dodge=dodge)
        df_agg = self._aggregate(df, _splitby, self._value)
        _pos_map = self._cat_iter.prep_position_map(_splitby, dodge=_dodge)

        xj = _jitter.CategoricalJitter(_splitby, _pos_map)
        yj = _jitter.IdentityJitter(self._value).check(df_agg)
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFLines.from_table(
            df_agg, xj, yj, name=name, color=color, width=width, style=style,
            backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(color, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        if alpha is not None:
            layer.update_alpha(alpha)
        return canvas.add_layer(layer)

    def add_markers(
        self,
        *,
        name: str | None = None,
        color: NStr | ColorType | None = None,
        alpha: float | None = None,
        hatch: NStr | Hatch | None = None,
        size: str | float | None = None,
        symbol: NStr | Symbol | None = None,
        dodge: NStr | bool = False,
    ) -> _lt.DFMarkers[_DF]:
        """
        Add markers that represent the aggregated values.

        >>> canvas.cat_x(df, "time", "value").mean().add_markers()

        Parameters
        ----------
        x : str
            Column name for x-axis.
        y : str
            Column name for y-axis.
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.

        Returns
        -------
        DFMarkers
            Marker collection layer.
        """
        canvas = self._canvas()
        df = self._df
        _splitby, _dodge = _shared.norm_dodge(
            df, self._offset, color, hatch, symbol, dodge=dodge
        )  # fmt: skip
        df_agg = self._aggregate(df, _splitby, self._value)
        _pos_map = self._cat_iter.prep_position_map(_splitby, dodge=_dodge)

        xj = _jitter.CategoricalJitter(_splitby, _pos_map)
        yj = _jitter.IdentityJitter(self._value).check(df_agg)
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFMarkers(
            df_agg, xj, yj, name=name, color=color, hatch=hatch, size=size,
            symbol=symbol, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(color, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        if alpha is not None:
            layer.update_alpha(alpha)
        return canvas.add_layer(layer)

    def add_bars(
        self,
        *,
        name: str | None = None,
        color: NStr | ColorType | None = None,
        alpha: float | None = None,
        hatch: NStr | Hatch | None = None,
        extent: float = 0.8,
        dodge: NStr | bool = True,
    ) -> _lt.DFBars[_DF]:
        """
        Add bars that represent the aggregated values.

        >>> canvas.cat_x(df, "time", "value").mean().add_bars()

        Parameters
        ----------
        x : str
            Column name for x-axis.
        y : str
            Column name for y-axis.
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        width : str, optional
            Column name for bar width. Must be numerical.

        Returns
        -------
        DFBars
            Bar collection layer.
        """
        canvas = self._canvas()
        df = self._df
        _splitby, _dodge = _shared.norm_dodge(
            df, self._offset, color, hatch, dodge=dodge
        )  # fmt: skip
        df_agg = self._aggregate(df, _splitby, self._value)
        _pos_map = self._cat_iter.prep_position_map(_splitby, dodge=_dodge)

        xj = _jitter.CategoricalJitter(_splitby, _pos_map)
        yj = _jitter.IdentityJitter(self._value).check(df_agg)

        _extent = self._cat_iter.zoom_factor(_dodge) * extent
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFBars.from_table(
            df_agg, xj, yj, name=name, color=color, hatch=hatch, extent=_extent,
            backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(color, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        if alpha is not None:
            layer.update_alpha(alpha)
        return canvas.add_layer(layer)

    def _aggregate(
        self,
        df: DataFrameWrapper[_DF],
        by: tuple[str, ...],
        on: str,
    ) -> DataFrameWrapper[_DF]:
        if self._agg_method == "size":
            return df.value_count(by)
        elif self._agg_method == "first":
            return df.value_first(by, on)
        else:
            if on is None:
                raise ValueError("Value column is not specified.")
            return df.agg_by(by, [on], self._agg_method)


class XCatPlotter(OneAxisCatPlotter[_C, _DF]):
    _orient = Orientation.VERTICAL


class YCatPlotter(OneAxisCatPlotter[_C, _DF]):
    _orient = Orientation.HORIZONTAL



from __future__ import annotations

from contextlib import contextmanager
from typing import Any, Generator

from whitecanvas.theme._dataclasses import DARK_THEME, LIGHT_THEME, Theme

_EXISTING_THEMES = {
    "light": LIGHT_THEME,
    "dark": DARK_THEME,
}
_DEFAULT_THEME = _EXISTING_THEMES["light"]


def get_theme(name: str | Theme | None = None) -> Theme:
    """
    Get a theme by name.

    >>> import whitecanvas as wc
    >>> theme = wc.theme.get_theme()  # get the default theme
    >>> theme = wc.theme.get_theme("dark")  # get the dark theme
    """
    if isinstance(name, Theme):
        return name
    if name is None:
        return _DEFAULT_THEME
    return _EXISTING_THEMES[name]


def _default(attr: str, value) -> Any:
    if value is not None:
        return value
    out = _DEFAULT_THEME
    for a in attr.split("."):
        out = getattr(out, a)
    return out


@contextmanager
def context(name: str | Theme | None = None) -> Generator[Theme, None, None]:
    """
    Temporarily change the default theme in this context.

    >>> import whitecanvas as wc
    >>> with wc.theme.context("dark") as theme:
    ...     theme.font.color = "yellow"  # change the theme parameters
    ...     # do something with the dark theme
    >>> # back to the default theme
    """
    global _DEFAULT_THEME
    theme = get_theme(name).copy()
    _default_theme = _DEFAULT_THEME
    theme._set_mutable(True)
    _DEFAULT_THEME = theme
    try:
        yield theme
    finally:
        _DEFAULT_THEME = _default_theme


def update_default(name: str | Theme | None = None) -> _UpdateDefault:
    """
    Update the default theme.

    This function can be directly called to update the default theme by name.

    >>> import whitecanvas as wc
    >>> wc.theme.update_default("dark")  # change the default theme

    You can also use it as a context manager to further change the theme parameters.
    The updated parameters persist after the context exits.

    >>> with wc.theme.update_default("dark") as theme:
    ...     theme.font.color = "yellow"  # change the theme parameters
    >>> # the default theme is now the dark theme with yellow font
    """
    return _UpdateDefault(name)


class _UpdateDefault:
    def __init__(self, name: str | Theme | None = None):
        global _DEFAULT_THEME
        self._theme = get_theme(name)
        self._default_theme = _DEFAULT_THEME.copy()
        _DEFAULT_THEME = self._theme

    def __enter__(self):
        self._theme._set_mutable(True)
        return self._theme

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _DEFAULT_THEME
        if exc_type is not None:
            _DEFAULT_THEME = self._default_theme
        self._theme._set_mutable(False)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from matplotlib.quiver import Quiver
from mpl_toolkits.mplot3d import proj3d
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import FakeAxes, MplLayer
from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from whitecanvas.backend.matplotlib.components3d import Canvas3D


class Vectors3D(Quiver, MplLayer):
    def __init__(self, x0, dx, y0, dy, z0, dz):
        super().__init__(
            FakeAxes(), x0, y0, dx, dy, angles="xy", scale_units="xy", scale=1
        )
        self._3d_x = x0
        self._3d_y = y0
        self._3d_z = z0
        self._3d_u = dx
        self._3d_v = dy
        self._3d_w = dz

    def do_3d_projection(self, renderer=None):
        x0, y0, z0 = proj3d.proj_transform(
            self._3d_x,
            self._3d_y,
            self._3d_z,
            self.axes.M,
        )
        x1, y1, z1 = proj3d.proj_transform(
            self._3d_x + self._3d_u,
            self._3d_y + self._3d_v,
            self._3d_z + self._3d_w,
            self.axes.M,
        )
        self.set_UVC(x1 - x0, y1 - y0)
        self.set_offsets(np.column_stack([x0, y0]))
        return np.min(z0)

    def post_add(self, canvas: Canvas3D):
        self.transform = canvas._axes.transData
        self.set_offset_transform(canvas._axes.transData)
        self._axes = canvas._axes

    def _plt_get_data(self):
        return self._3d_x, self._3d_y, self._3d_z, self._3d_u, self._3d_v, self._3d_w

    def _plt_set_data(self, x0, dx, y0, dy, z0, dz):
        self._3d_x = x0
        self._3d_y = y0
        self._3d_z = z0
        self._3d_u = dx
        self._3d_v = dy
        self._3d_w = dz
        self.do_3d_projection()

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float):
        self.set_linewidth(width)

    def _plt_get_edge_style(self) -> LineStyle:
        return LineStyle(self.get_linestyle())

    def _plt_set_edge_style(self, style: LineStyle):
        self.set_linestyle(style.value)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_edgecolor()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_color(color)

    def _plt_get_antialias(self) -> bool:
        return self.get_antialiased()

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)



import numpy as np
from scipy import ndimage as ndi

from whitecanvas import vgrid_nonuniform


def main():
    grid = vgrid_nonuniform(heights=[3, 1], backend="pyqtgraph:qt")
    c0 = grid.add_canvas(0)
    c1 = grid.add_canvas(1)
    image = np.random.random((100,100))
    c0.add_image(image)
    p0 = [20, 80]
    p1 = [50, 60]

    def get_profile(p0, p1):
        length = np.sqrt((p0[0] - p1[0])** 2 + (p0[1] - p1[1])** 2)
        xs = np.linspace(p0[1], p1[1], int(length))
        ys = np.linspace(p0[0], p1[0], int(length))
        return ndi.map_coordinates(image, np.stack([ys, xs], axis=0))

    c0.add_line([p0[1], p1[1]], [p0[0], p1[0]], color="yellow", width=3)
    c1.add_line(get_profile(p0, p1), color="black")
    grid.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

from contextlib import contextmanager

_INSTALLED_MODULES = {}

_DEFAULT_APP = {
    "pyqtgraph": "qt",
    "vispy": "qt",
}


class Backend:
    """The backend of plots."""

    _default: str = "matplotlib"

    def __init__(self, name: Backend | str | None = None) -> None:
        if name is None:
            name = self._default
        if isinstance(name, Backend):
            app = name._app
            name = name._name
        elif isinstance(name, str):
            if ":" in name:
                name, app = name.split(":", maxsplit=1)
                # normalize app name
                if app == "nb":
                    app = "notebook"
            else:
                app = _DEFAULT_APP.get(name, "default")
        else:
            raise TypeError(f"Backend name must be str or Backend, not {type(name)}")
        if name in _INSTALLED_MODULES:
            self._mod = _INSTALLED_MODULES[name]
        else:
            from importlib import import_module

            _INSTALLED_MODULES[name] = self._mod = import_module(
                f"whitecanvas.backend.{name}"
            )
        self._name: str = name
        self._app: str = app
        self.__class__._default = name

    def __repr__(self) -> str:
        return f"<Backend {self._name!r} (app: {self._app!r})>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Backend):
            return False
        return self._name == other._name and self._app == other._app

    @property
    def name(self) -> str:
        """Name of the backend."""
        return self._name

    @property
    def app(self) -> str:
        """Name of the application."""
        return self._app

    def has(self, attr: str) -> bool:
        """Check if the current backend has an object."""
        return hasattr(self._mod, attr)

    def get(self, attr: str):
        """Get an object from the current backend."""
        out = getattr(self._mod, attr, None)
        if out is None:
            raise RuntimeError(f"Backend {self._name!r} does not implement {attr!r}")
        return out

    def get_submodule(self, attr: str):
        """Get a submodule from the current backend."""
        import importlib

        if self.is_dummy():
            return _dummy_backend_module
        out = importlib.import_module(f"whitecanvas.backend.{self._name}.{attr}")
        if out is None:
            raise RuntimeError(f"Backend {self._name!r} does not implement {attr!r}")
        return out

    def is_dummy(self) -> bool:
        """True is the backend is a dummy backend."""
        return self.name.startswith(".")


class DummyObject:
    def __init__(self, *args, **kwargs):
        pass


class DummyBackendModule:
    def __init__(self):
        self.Canvas = DummyObject

    def __getattr__(self, key):
        return DummyObject


_dummy_backend_module = DummyBackendModule()


@contextmanager
def patch_dummy_backend():
    dummy_name = "."
    while dummy_name in _INSTALLED_MODULES:
        dummy_name += "."
    _INSTALLED_MODULES[dummy_name] = _dummy_backend_module
    try:
        yield dummy_name
    finally:
        del _INSTALLED_MODULES[dummy_name]



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.markers as mmarkers

if TYPE_CHECKING:
    from matplotlib.backend_bases import MouseEvent as mplMouseEvent

    from whitecanvas.backend.matplotlib.canvas import Canvas
    from whitecanvas.types import Symbol


class MplLayer:
    def _plt_get_visible(self):
        return self.get_visible()

    def _plt_set_visible(self, visible):
        self.set_visible(visible)

    def _plt_set_zorder(self, zorder: int):
        self.set_zorder(zorder)


class FakeAxes:
    def __init__(self):
        self.transData = None
        self.transAxes = None


def symbol_to_path(symbol: Symbol):
    marker_obj = mmarkers.MarkerStyle(symbol.value)
    return marker_obj.get_path().transformed(marker_obj.get_transform())


OVERLAY_ZORDER = 10000


def as_overlay(layer: MplLayer, canvas):
    layer.set_transform(canvas._axes.transAxes)
    layer.set_zorder(OVERLAY_ZORDER)


class MplMouseEventsMixin(MplLayer):
    def __init__(self):
        self._pick_callbacks = []
        self._hover_texts: list[str] | None = None

    def _on_hover(self, event: mplMouseEvent = None):
        if self._hover_texts is None or not self._plt_get_visible():
            return
        contains, ind = self.contains(event)
        if not contains:
            return
        indices = ind["ind"]
        if len(indices) == 0:
            return
        index: int = indices[0]
        hover_text = self._hover_texts[index]
        return hover_text

    def _plt_set_hover_text(self, texts: list[str]):
        self._hover_texts = texts

    def post_add(self, canvas: Canvas):
        fig = canvas._axes.figure
        if fig is None:
            return
        for cb in self._pick_callbacks:
            fig.canvas.mpl_connect("pick_event", cb)

    def _plt_connect_pick_event(self, callback):
        def cb(event):
            if event.artist is not self:
                return
            callback(event.ind)

        self._pick_callbacks.append(cb)



from __future__ import annotations

from typing import Any, Mapping

import numpy as np
from cmap import Color
from numpy.typing import ArrayLike, NDArray

from whitecanvas.types import XYData, XYZData
from whitecanvas.utils import type_check as _tc


def as_array_1d(x: ArrayLike, dtype=None) -> NDArray[np.number]:
    """Normalize the input as a 1D array."""
    x = np.asarray(x, dtype=dtype)
    if x.ndim != 1:
        raise ValueError(f"Expected 1D array, got {x.ndim}D array")
    if x.dtype.kind not in "iuf":
        raise ValueError(f"Input {x!r} did not return a numeric array")
    return x


def normalize_xy(*args) -> tuple[NDArray[np.number], NDArray[np.number]]:
    """Normalize the input as two 1D array with the same shape."""
    if len(args) == 1:
        if isinstance(args[0], XYData):
            return args[0].x, args[0].y
        arr = np.asarray(args[0])
        if arr.dtype.kind not in "iuf":
            raise ValueError(f"Input {args[0]!r} did not return a numeric array")
        if arr.ndim == 1:
            ydata = arr
            xdata = np.arange(ydata.size)
        elif arr.ndim == 2 and arr.shape[1] == 2:
            xdata = arr[:, 0]
            ydata = arr[:, 1]
        else:
            raise ValueError(
                f"Expected 1D array or 2D array with shape (N, 2), got {arr.shape}"
            )
    elif len(args) == 2:
        if np.isscalar(args[0]) and np.isscalar(args[1]):
            return np.array(args[0]), np.array(args[1])
        xdata = as_array_1d(args[0])
        if not hasattr(args[1], "__array__") and callable(args[1]):
            ydata = args[1](xdata)
        else:
            ydata = as_array_1d(args[1])
        if xdata.size != ydata.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {xdata.size} and {ydata.size}"
            )
    else:
        raise TypeError(f"Expected 1 or 2 positional arguments, got {len(args)}")
    return xdata, ydata


def normalize_xyz(
    *args,
) -> tuple[NDArray[np.number], NDArray[np.number], NDArray[np.number]]:
    """Normalize the input as three 1D array with the same shape."""
    if len(args) == 1:
        if isinstance(args[0], XYZData):
            return args[0].x, args[0].y, args[0].z
        arr = np.asarray(args[0])
        if arr.dtype.kind not in "iuf":
            raise ValueError(f"Input {args[0]!r} did not return a numeric array")
        if arr.ndim == 2 and arr.shape[1] == 3:
            xdata = arr[:, 0]
            ydata = arr[:, 1]
            zdata = arr[:, 2]
        else:
            raise ValueError(
                f"Expected 1D array or 2D array with shape (N, 3), got {arr.shape}"
            )
    elif len(args) == 3:
        xdata = as_array_1d(args[0])
        ydata = as_array_1d(args[1])
        zdata = as_array_1d(args[2])
        if xdata.size != ydata.size != zdata.size:
            raise ValueError(
                "Expected xdata, ydata, and zdata to have the same size, "
                f"got {xdata.size}, {ydata.size}, and {zdata.size}"
            )
    else:
        raise TypeError(f"Expected 1 or 3 positional arguments, got {len(args)}")
    return xdata, ydata, zdata


def arr_color(color) -> np.ndarray:
    """Normalize a color input to a 4-element float array."""
    try:
        c = Color(color)
    except Exception:
        raise ValueError(f"Invalid input for a color: {color!r}") from None
    return np.array(c.rgba, dtype=np.float32)


def hex_color(color) -> str:
    """Normalize a color input to a #RRGGBBAA string."""
    return Color(color).hex


def rgba_str_color(color) -> str:
    """Normalize a color input to a rgba(r, g, b, a) string."""
    return Color(color).rgba_string


def as_any_1d_array(x: Any, size: int, dtype=None) -> np.ndarray:
    if _tc.is_not_array(x):
        out = np.full((size,), x, dtype=dtype)
    else:
        out = np.asarray(x, dtype=dtype)
        if out.shape != (size,):
            raise ValueError(f"Expected shape ({size},), got {out.shape}")
    return out


def as_color_array(color, size: int) -> NDArray[np.float32]:
    if isinstance(color, str):  # e.g. color = "black"
        col = arr_color(color)
        return np.repeat(col[np.newaxis, :], size, axis=0)
    if isinstance(color, np.ndarray):
        if color.size == 0 and size == 0:
            return color
        if color.dtype.kind in "OU":
            if color.shape != (size,):
                raise ValueError(
                    f"Expected color array of shape ({size},), got {color.shape}"
                )
            return np.stack([arr_color(each) for each in color], axis=0)
        elif color.shape in [(3,), (4,)]:
            col = arr_color(color)
            return np.repeat(col[np.newaxis, :], size, axis=0)
        elif color.shape in [(size, 3), (size, 4)]:
            return color
        else:
            if isinstance(color, np.ndarray):
                color_repr = f"ndarray of shape={color.shape} and dtype={color.dtype}"
            else:
                color_repr = repr(color)
            raise ValueError(
                f"Color array must have shape (3,), (4,), (N={size}, 3), or (N={size},"
                f" 4) but got {color_repr}."
            )
    try:
        arr = np.array(color)
    except ValueError as e:
        if str(e).startswith("setting an array element with a sequence"):
            # this happens when ["red", [0.0, 1.0, 0.0, 1.0]] is given
            color_normed = np.stack(
                [Color(c).rgba for c in color],
                axis=0,
                dtype=np.float32,
            )
            return color_normed
        else:
            raise
    return as_color_array(arr, size)


def parse_texts(template: str, ndata: int, extra: Any | None = None) -> dict[str, Any]:
    """Parse a template string and return parameters of hover texts."""
    params = {}
    if extra is None:
        extra = {}
    elif isinstance(extra, Mapping):
        pass
    elif _tc.is_pandas_dataframe(extra):
        return parse_texts(template, ndata, extra=extra.to_dict(orient="list"))
    elif _tc.is_polars_dataframe(extra):
        return parse_texts(template, ndata, extra=extra.to_dict())
    else:
        raise TypeError("extra must be a mapping.")
    for k, v in extra.items():
        if "{" + k not in template:
            continue
        if _tc.is_not_array(v):
            _v = np.full(ndata, v)
        else:
            _v = as_any_1d_array(v, ndata)
        if _v.size != ndata:
            raise ValueError(f"Expected {_v.size} elements, got {ndata} for {k!r}.")
        params[k] = v
    return params



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import BarProtocol, check_protocol
from whitecanvas.utils.normalize import as_color_array


@check_protocol(BarProtocol)
class Bars(visuals.Compound):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        center = np.stack([(xlow + xhigh) / 2, (ylow + yhigh) / 2], axis=1)
        rectangles: list[visuals.Rectangle] = []
        for c, w, h in zip(center, xhigh - xlow, yhigh - ylow):
            _w = abs(w)
            _h = abs(h)
            if _w == 0:
                _w = _h * 1e-6
            elif _h == 0:
                _h = _w * 1e-6
            rectangles.append(
                visuals.Rectangle(center=c, width=_w, height=_h, border_width=0)
            )
        super().__init__(rectangles)
        self.unfreeze()
        self._rectangles = rectangles

    ##### LayerProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        data = []
        for rect in self._rectangles:
            xc, yc = rect.center
            w = rect.width
            h = rect.height
            data.append([xc - w / 2, xc + w / 2, yc - h / 2, yc + h / 2])
        return tuple(np.array(data).T)

    def _plt_set_data(self, xlow, xhigh, ylow, yhigh):
        center = np.stack([(xlow + xhigh) / 2, (ylow + yhigh) / 2], axis=1)
        for rect, c, w, h in zip(self._rectangles, center, xhigh - xlow, yhigh - ylow):
            rect.center = c
            rect.width = max(abs(w), 1e-8)
            rect.height = max(abs(h), 1e-8)

    def _plt_get_ndata(self):
        return len(self._rectangles)

    ##### HasFace protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        colors = []
        for rect in self._rectangles:
            colors.append(rect.color)
        return np.stack(colors, axis=0)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, size=self._plt_get_ndata())
        for rect, c in zip(self._rectangles, color):
            rect.color = c

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()

    ##### HasEdges protocol #####
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        colors = []
        for rect in self._rectangles:
            colors.append(rect.border_color)
        return np.stack(colors, axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, size=self._plt_get_ndata())
        for rect, c in zip(self._rectangles, color):
            rect.border_color = c

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        # NOTE: vispy forgot to add `border_width` to `Rectangle` class?
        widths = []
        for rect in self._rectangles:
            widths.append(rect._border_width)
        return np.array(widths)

    def _plt_set_edge_width(self, width: float):
        if np.isscalar(width):
            width = np.full(self._plt_get_ndata(), width)
        for rect, w in zip(self._rectangles, width):
            rect._border_width = w

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _plt_connect_pick_event(self, callback):
        pass



from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(visuals.Line):
    def __init__(self, xdata, ydata):
        data = np.stack([xdata, ydata], axis=1)
        super().__init__(data, antialias=True)
        self.unfreeze()

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        pos = self.pos
        return pos[:, 0], pos[:, 1]

    def _plt_set_data(self, xdata, ydata):
        self.set_data(np.stack([xdata, ydata], axis=1))

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.width

    def _plt_set_edge_width(self, width: float):
        self.set_data(width=width)

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_style()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.color

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_data(color=color)

    def _plt_get_antialias(self) -> bool:
        return self.antialias

    def _plt_set_antialias(self, antialias: bool):
        self.antialias = antialias

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _plt_connect_pick_event(self, callback):
        pass


def _make_connection(data: NDArray[np.number]):
    if len(data) == 0:
        return np.empty((0, 2), dtype=np.uint32)
    connections = []
    npoints = 0
    for seg in data:
        conn = np.column_stack(
            [
                np.arange(npoints, npoints + len(seg) - 1, dtype=np.uint32),
                np.arange(npoints + 1, npoints + len(seg), dtype=np.uint32),
            ]
        )
        connections.append(conn)
        npoints += len(seg)
    return np.concatenate(connections)


def _safe_concat(data: list[NDArray[np.floating]]):
    if len(data) == 0:
        return None
    return np.concatenate(data, axis=0)


@check_protocol(MultiLineProtocol)
class MultiLine(visuals.Line):
    def __init__(self, data: list[NDArray[np.float32]]):
        self._data_raw = data
        self._seg_colors = np.ones((len(data), 4), dtype=np.float32)
        super().__init__(
            pos=_safe_concat(data),
            antialias=True,
            connect=_make_connection(data),
        )
        self.unfreeze()

    def _plt_get_ndata(self):
        return len(self._data_raw)

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self._data_raw

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        connections = _make_connection(data)
        self._data_raw = data
        self.set_data(pos=_safe_concat(data), connect=connections)

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        ndata = len(self._data_raw)
        return np.full(ndata, self.width, dtype=np.float32)

    def _plt_set_edge_width(self, width):
        if is_real_number(width):
            self.set_data(width=width)
        else:
            candidates = np.unique(width)
            if len(candidates) == 1:
                self.set_data(width=width[0])
            elif len(candidates) == 0:
                return
            else:
                warnings.warn(
                    "vispy backend does not support different edge width for each line "
                    "segment. The first value is used for all line",
                    UserWarning,
                    stacklevel=2,
                )
                self.set_data(width=width[0])

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self._seg_colors

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        ndata = len(self._data_raw)
        colors = as_color_array(color, ndata)
        raw_colors = []
        for color, seg in zip(colors, self._data_raw):
            raw_colors.append(np.repeat(color[np.newaxis], len(seg), axis=0))
        self.set_data(color=_safe_concat(raw_colors))
        self._seg_colors = colors

    def _plt_get_antialias(self) -> bool:
        return self.antialias

    def _plt_set_antialias(self, antialias: bool):
        self.antialias = antialias

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _plt_connect_pick_event(self, callback):
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, TypeVar

import numpy as np
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._base import HoverableDataBoundLayer
from whitecanvas.layers._mixin import (
    ConstEdge,
    ConstFace,
    EdgeNamespace,
    FaceEdgeMixinEvents,
    FaceNamespace,
    MultiEdge,
    MultiFace,
    MultiFaceEdgeMixin,
)
from whitecanvas.protocols import BarProtocol
from whitecanvas.types import ColorType, Hatch, LineStyle, Rect, _Void
from whitecanvas.utils.normalize import parse_texts

if TYPE_CHECKING:
    from typing_extensions import Self

_void = _Void()
_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)


class RectsEvents(FaceEdgeMixinEvents):
    clicked = Signal(int)


class Rects(
    HoverableDataBoundLayer[BarProtocol, NDArray[np.float32]],
    MultiFaceEdgeMixin[_Face, _Edge],
):
    """
    Layer that represents Rectangles.

    Attributes
    ----------
    face : `whitecanvas.layers._mixin.FaceNamespace`
        Face properties of the bars.
    edge : `whitecanvas.layers._mixin.EdgeNamespace`
        Edge properties of the bars.
    """

    _backend_class_name = "Bars"
    events: RectsEvents
    _events_class = RectsEvents

    def __init__(
        self,
        coords: NDArray[np.number],
        *,
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        MultiFaceEdgeMixin.__init__(self)
        xxyy, xhint, yhint = _norm_inputs(coords)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), *xxyy)
        self.face.update(color=color, alpha=alpha, hatch=hatch)
        self._x_hint, self._y_hint = xhint, yhint
        self._init_events()
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _get_layer_data(self) -> NDArray[np.float32]:
        """Current data of the layer."""
        x0, x1, y0, y1 = self._backend._plt_get_data()
        return np.column_stack([x0, x1, y0, y1])

    def _norm_layer_data(self, data: ArrayLike) -> NDArray[np.float32]:
        arr = np.atleast_2d(data).astype(np.float32, copy=False)
        if arr.size > 0 and arr.shape[1] != 4:
            raise ValueError("Data must have 4 columns")
        return arr

    def _set_layer_data(self, data: NDArray[np.float32]):
        self._backend._plt_set_data(*data.T)
        _, self._x_hint, self._y_hint = _norm_inputs(data)

    @property
    def rects(self) -> list[Rect]:
        """Return the data as a list of `Rect` objects."""
        return [Rect.with_sort(*xy) for xy in self._get_layer_data()]

    @property
    def ndata(self) -> int:
        """The number of data points"""
        return self.data.shape[0]

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Rects[ConstFace, _Edge]:
        return super().with_face(color=color, hatch=hatch, alpha=alpha)

    def with_face_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] | _Void = _void,
        alpha: float = 1,
    ) -> Rects[MultiFace, _Edge]:
        return super().with_face_multi(color=color, hatch=hatch, alpha=alpha)

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Rects[_Face, ConstEdge]:
        return super().with_edge(color=color, width=width, style=style, alpha=alpha)

    def with_edge_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Rects[_Face, MultiEdge]:
        return super().with_edge_multi(
            color=color, width=width, style=style, alpha=alpha
        )

    def as_edge_only(
        self,
        *,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the rectangles to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        color = self.face.color
        if color.ndim == 0:
            pass
        elif color.ndim == 1:
            self.with_edge(color=color, width=width, style=style)
        elif color.ndim == 2:
            self.with_edge_multi(color=color, width=width, style=style)
        else:
            raise RuntimeError("Unreachable error.")
        self.face.update(alpha=0.0)
        return self

    def with_hover_template(self, template: str, extra: Any | None = None) -> Rects:
        coords = self.data
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, coords.shape[0], extra)
        # set default format keys
        params.setdefault("left", coords[:, 0])
        params.setdefault("right", coords[:, 1])
        params.setdefault("bottom", coords[:, 2])
        params.setdefault("top", coords[:, 3])
        if "i" not in params:
            params["i"] = np.arange(coords.shape[0])
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(coords.shape[0])
        ]
        self._backend._plt_set_hover_text(texts)
        return self

    def _as_legend_item(self) -> _legend.BarLegendItem:
        return _legend.BarLegendItem(
            self.face._as_legend_info(), self.edge._as_legend_info()
        )


def _norm_inputs(coords: ArrayLike):
    arr = np.atleast_2d(coords).astype(np.float32, copy=False)
    if arr.size == 0:
        return (np.zeros(0, dtype=np.float32),) * 4, None, None
    if arr.shape[1] != 4:
        raise ValueError("Data must have 4 columns")
    x0, x1, y0, y1 = arr.T
    if x0.size > 0:
        xhint = np.min(x0), np.max(x1)
        yhint = np.min(y0), np.max(y1)
    else:
        xhint = None
        yhint = None
    return (x0, x1, y0, y1), xhint, yhint



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, TypeVar, overload

from psygnal import Signal, SignalGroup
from psygnal.containers import EventedList

from whitecanvas.layers import Layer, LayerGroup, PrimitiveLayer

if TYPE_CHECKING:
    from typing_extensions import Self

_V = TypeVar("_V", bound=Any)


class LayerListEvents(SignalGroup):
    inserting = Signal(int)  # idx
    inserted = Signal(int, Layer)  # (idx, value)
    removing = Signal(int)  # idx
    removed = Signal(int, Layer)  # (idx, value)
    moving = Signal(int, int)  # (src_idx, dest_idx)
    moved = Signal(int, int, Layer)  # (src_idx, dest_idx, value)
    changed = Signal(object, Layer, Layer)  # (int | slice, old, new)
    reordered = Signal()
    renamed = Signal(int, str, str)  # (idx, old_name, new_name)


class LayerList(EventedList[Layer]):
    events: LayerListEvents

    def __init__(self, data: Iterable[Layer] = ()):
        super().__init__(data, hashable=True, child_events=False)
        self._instances: dict[int, Self] = {}

    def __get__(self, instance, owner) -> Self:
        if instance is None:
            return self
        _id = id(instance)
        cls = self.__class__
        if (out := self._instances.get(_id, None)) is None:
            out = self._instances[_id] = cls()
        return out

    @overload
    def __getitem__(self, idx: int) -> Layer:
        ...

    @overload
    def __getitem__(self, idx: slice) -> Self:
        ...

    @overload
    def __getitem__(self, idx: str) -> Layer:
        ...

    def __getitem__(self, idx):
        if isinstance(idx, str):
            for layer in self:
                if layer.name == idx:
                    return layer
            raise KeyError(idx)
        return super().__getitem__(idx)

    def get(self, idx: str, default: _V | None = None) -> Layer | _V | None:
        if isinstance(idx, str):
            for layer in self:
                if layer.name == idx:
                    return layer
            return default
        raise TypeError(f"LayerList.get() expected str, got {type(idx)}")

    def iter_primitives(self) -> Iterable[PrimitiveLayer]:
        for layer in self:
            if isinstance(layer, LayerGroup):
                yield from layer.iter_primitive()
            else:
                yield layer



from __future__ import annotations

from typing import Any

import numpy as np
from numpy.typing import NDArray

from whitecanvas.layers import _legend, _text_utils
from whitecanvas.layers._primitive import Markers, MultiLine, Texts
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.layers.group._offsets import NoOffset, TextOffset
from whitecanvas.types import Alignment, ColorType


class Graph(LayerContainer):
    def __init__(
        self,
        nodes: Markers,
        edges: MultiLine,
        texts: Texts,
        edges_data: NDArray[np.intp],
        name: str | None = None,
        offset: TextOffset | None = None,
    ):
        if offset is None:
            offset = NoOffset()
        self._edges_data = edges_data
        super().__init__([nodes, edges, texts], name=name)
        self._text_offset = offset

    def _default_ordering(self, n: int) -> list[int]:
        assert n == 3
        return [1, 0, 2]

    @property
    def nodes(self) -> Markers:
        """The nodes layer."""
        return self._children[0]

    @property
    def edges(self) -> MultiLine:
        """The edges layer."""
        return self._children[1]

    @property
    def texts(self) -> Texts:
        """The texts layer."""
        return self._children[2]

    @property
    def edge_indices(self) -> NDArray[np.intp]:
        """Current data of the edges."""
        return self._edges_data

    @property
    def text_offset(self) -> TextOffset:
        """Return the text offset."""
        return self._text_offset

    def with_text_offset(self, dx: Any, dy: Any):
        """Add offset to text positions."""
        _offset = self._text_offset._add(dx, dy)
        if self.texts.ndata > 0:
            data = self.nodes.data
            xoff, yoff = _offset._asarray()
            self.texts.set_pos(data.x + xoff, data.y + yoff)
        self._text_offset = _offset
        return self

    add_text_offset = with_text_offset

    def set_graph(self, nodes: NDArray[np.floating], edges: NDArray[np.intp]):
        """Set the graph data."""
        self.nodes.data = nodes
        self._edges_data = edges

    def with_text(
        self,
        strings: list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.CENTER,
        fontfamily: str | None = None,
        offset: tuple[Any, Any] | None = None,
    ):
        if isinstance(strings, str):
            strings = [strings] * self.nodes.data.x.size
        if offset is None:
            _offset = self._text_offset
        else:
            _offset = NoOffset()._add(*offset)

        strings = _text_utils.norm_label_text(strings, self.nodes.data)
        xdata, ydata = self.nodes.data
        dx, dy = _offset._asarray()
        self.texts.string = strings
        self.texts.set_pos(xdata + dx, ydata + dy)
        self.texts.update(
            color=color,
            size=size,
            rotation=rotation,
            anchor=anchor,
            family=fontfamily,
        )
        return self

    def _as_legend_item(self):
        line = self.edges._as_legend_item()
        markers = self.nodes._as_legend_item()
        return _legend.PlotLegendItem(line, markers)



# This example shows how to efficiently use scipy.optimize.curve_fit

import numpy as np
from scipy.optimize import curve_fit

from whitecanvas import new_canvas


def sample_data(tau: float, a: float, b: float, size: int = 40):
    x = np.arange(size)
    y = a * np.exp(-x / tau) + np.random.normal(size=size, scale=a*0.1) + b
    return x, y

def main():
    np.random.seed(1462)
    canvas = new_canvas(backend="matplotlib:qt")

    # tau, a, b, size
    params_true = [
        (9.1, 3.6, 0.46, 40),
        (6.8, 3.0, 0.21, 48),
        (7.6, 4.0, 0.58, 32)
    ]

    # fitting model
    def model(x, a, tau, b):
        return a * np.exp(-x / tau) + b

    for i, p in enumerate(params_true):
        # add raw data
        x, y = sample_data(*p)
        markers = canvas.add_markers(x, y, alpha=0.25, name=f"data-{i}")

        # add the fitting curve
        params, _ = curve_fit(model, x, y, p0=[2, 1, 0])
        (
            canvas
            .add_infcurve(model, color=markers.face.color, name=f"fit of data-{i}")
            .update_params(*params)
            .with_hover_text("a={:.3g}, tau={:.3g}, b={:.3g}".format(*params))
        )

    canvas.add_legend()
    canvas.update_labels(x="time [sec]", y="intensity [a.u.]", title="Fitting results")
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from whitecanvas.types._alias import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    HistBinType,
    KdeBandWidthType,
    _Void,
)
from whitecanvas.types._enums import (
    Alignment,
    Hatch,
    HistogramKind,
    HistogramShape,
    LineStyle,
    Location,
    LocationStr,
    Modifier,
    MouseButton,
    MouseEventType,
    Orientation,
    OrientationLike,
    Origin,
    StepStyle,
    StepStyleStr,
    Symbol,
)
from whitecanvas.types._mouse import MouseEvent, Point
from whitecanvas.types._tuples import (
    MeshData,
    Rect,
    XYData,
    XYTextData,
    XYVectorData,
    XYYData,
    XYZData,
    XYZVectorData,
)

__all__ = [
    "ColorType",
    "ColormapType",
    "ArrayLike1D",
    "LineStyle",
    "Location",
    "LocationStr",
    "Symbol",
    "Hatch",
    "HistBinType",
    "HistogramKind",
    "HistogramShape",
    "KdeBandWidthType",
    "Orientation",
    "OrientationLike",
    "Origin",
    "Modifier",
    "MouseButton",
    "MouseEvent",
    "Point",
    "StepStyle",
    "StepStyleStr",
    "MouseEventType",
    "Alignment",
    "XYData",
    "XYYData",
    "XYTextData",
    "XYZData",
    "XYVectorData",
    "XYZVectorData",
    "Rect",
    "MeshData",
    "_Void",
]



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

from matplotlib import pyplot as plt
from matplotlib.ticker import AutoLocator, AutoMinorLocator

from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    from matplotlib import axis as mpl_axis
    from matplotlib import text as mpl_text

    from whitecanvas.backend.matplotlib.canvas import Canvas


class SupportsText:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)
        self._text = ""
        self._fontdict = {
            "color": plt.rcParams["text.color"],
            "fontsize": plt.rcParams["font.size"],
            "family": plt.rcParams["font.family"],
        }

    def _plt_get_color(self):
        return self._fontdict["color"]

    def _plt_get_size(self) -> str:
        return self._fontdict["fontsize"]

    def _plt_get_fontfamily(self) -> str:
        return self._fontdict["family"]


class Title(SupportsText):
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)
        self._text = ""
        self._fontdict = {
            "color": plt.rcParams["text.color"],
            "fontsize": plt.rcParams["font.size"],
            "family": plt.rcParams["font.family"],
        }

    def _plt_get_text(self) -> str:
        return self._canvas()._axes.get_title()

    def _plt_set_text(self, text: str):
        self._canvas()._axes.set_title(text, self._fontdict)
        self._text = text

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return bool(self._canvas()._axes.get_title())

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._canvas()._axes.set_title(self._text, fontdict=self._fontdict)
        else:
            self._canvas()._axes.set_title("")

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        self._canvas()._axes.set_title(self._text, fontdict=d)
        self._fontdict = d


class MplLabel(SupportsText):
    def __init__(self, canvas: Canvas, name: str):
        super().__init__(canvas)
        self._axis_name = name

    def _set_label(self, text: str, fontdict: dict[str, str] = {}):
        fname = f"set_{self._axis_name}label"
        getattr(self._canvas()._axes, fname)(text, fontdict)

    def _plt_get_text(self) -> str:
        return getattr(self._canvas()._axes, f"get_{self._axis_name}label")()

    def _plt_set_text(self, text: str):
        self._set_label(text, self._fontdict)
        self._text = text

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        self._set_label(self._text, fontdict=d)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        return bool(self._canvas()._axes.get_xlabel())

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._set_label(self._text, fontdict=self._fontdict)
        else:
            self._set_label("")

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        self._set_label(self._text, fontdict=d)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        self._set_label(self._text, fontdict=d)
        self._fontdict = d


class MplTicks(SupportsText):
    def __init__(self, canvas: Canvas, name: str):
        super().__init__(canvas)
        self._axis_name = name

    def _get_tick_labels(self) -> list[mpl_text.Text]:
        axes = self._canvas()._axes
        return getattr(axes, f"get_{self._axis_name}ticklabels")()

    def _get_ticks(self) -> list[float]:
        axes = self._canvas()._axes
        return getattr(axes, f"get_{self._axis_name}ticks")()

    def _set_ticks(self, ticks: list[float]):
        axes = self._canvas()._axes
        getattr(axes, f"set_{self._axis_name}ticks")(ticks)

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        return self._get_ticks(), [x.get_text() for x in self._get_tick_labels()]

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        axes = self._canvas()._axes
        getattr(axes, f"set_{self._axis_name}ticks")(pos, labels)

    def _plt_reset_override(self):
        # FIXME: This is not a perfect solution. It will update the x-axis scale.
        axis = getattr(self._canvas()._axes, f"get_{self._axis_name}axis")()
        axis.clear()

    def _plt_set_color(self, color):
        d = self._fontdict.copy()
        d["color"] = color
        for x in self._get_tick_labels():
            x.set_color(color)
        self._fontdict = d

    def _plt_get_visible(self) -> bool:
        ticklines = getattr(self._canvas()._axes, f"get_{self._axis_name}ticklines")()
        return ticklines[0].get_visible()

    def _plt_set_visible(self, visible: bool):
        axes = self._canvas()._axes
        for tick in getattr(axes, f"get_{self._axis_name}ticklines")():
            tick.set_visible(visible)
        for text in self._get_tick_labels():
            text.set_visible(visible)

    def _plt_set_size(self, size: str):
        d = self._fontdict.copy()
        d["fontsize"] = size
        for x in self._get_tick_labels():
            x.set_fontsize(size)
        self._fontdict = d

    def _plt_set_fontfamily(self, font):
        d = self._fontdict.copy()
        d["family"] = font
        for x in self._get_tick_labels():
            x.set_fontfamily(font)
        self._fontdict = d

    def _plt_get_text_rotation(self) -> float:
        return self._get_tick_labels()[0].get_rotation()

    def _plt_set_text_rotation(self, rotation: float):
        for x in self._get_tick_labels():
            x.set_rotation(rotation)


class MplAxis:
    def __init__(self, canvas: Canvas, name: str):
        self._canvas = weakref.ref(canvas)
        self._axis_name = name

    def _plt_get_limits(self) -> tuple[float, float]:
        axes = self._canvas()._axes
        x0, x1 = getattr(axes, f"get_{self._axis_name}lim")()
        if getattr(axes, f"{self._axis_name}axis_inverted")():
            return x1, x0
        else:
            return x0, x1

    def _plt_set_limits(self, limits: tuple[float, float]):
        axes = self._canvas()._axes
        if getattr(axes, f"{self._axis_name}axis_inverted")():
            limits = limits[::-1]
        getattr(axes, f"set_{self._axis_name}lim")(*limits)

    def _get_mpl_axis(self) -> mpl_axis.Axis:
        return getattr(self._canvas()._axes, f"{self._axis_name}axis")

    def _plt_get_color(self):
        return self._get_mpl_axis().get_tick_params()["color"]

    def _plt_set_color(self, color):
        ax = self._get_mpl_axis()
        color = tuple(color)
        ax.set_tick_params(color=color, labelcolor=color)
        if hasattr(ax, "line"):  # 3D
            ax.line.set_color(color)
        else:
            if self._axis_name == "x":
                self._canvas()._axes.spines["bottom"].set_color(color)
            else:
                self._canvas()._axes.spines["left"].set_color(color)

    def _plt_flip(self):
        axis = self._get_mpl_axis()
        axis.set_inverted(not axis.get_inverted())

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        axis = self._get_mpl_axis()
        axis.grid(
            visible,
            which="major",
            color=color,
            linestyle=style.value,
            linewidth=width,
        )



from __future__ import annotations

from typing import Sequence

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from pyqtgraph.GraphicsScene.mouseEvents import HoverEvent as pgHoverEvent
from qtpy import QtCore, QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_brush_style,
    from_qt_line_style,
    to_qt_brush_style,
    to_qt_line_style,
)
from whitecanvas.protocols import BarProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array


@check_protocol(BarProtocol)
class Bars(pg.BarGraphItem, PyQtLayer):
    clicked = QtCore.Signal(object)

    def __init__(self, xlow, xhigh, ylow, yhigh):
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        ndata = len(xlow)
        super().__init__(
            x0=xlow, x1=xhigh, y0=ylow, y1=yhigh,
            pens=[QtGui.QPen(pen) for _ in range(ndata)],
            brushes=[QtGui.QBrush(QtGui.QColor(0, 0, 0)) for _ in range(ndata)],
        )  # fmt: skip
        self._hover_texts: list[str] = None
        self._toolTipCleared = True

    ##### XYDataProtocol #####
    def _plt_get_data(self) -> Sequence[NDArray[np.floating]]:
        return self.opts["x0"], self.opts["x1"], self.opts["y0"], self.opts["y1"]

    def _plt_set_data(self, xlow, xhigh, ylow, yhigh):
        self.setOpts(x0=xlow, x1=xhigh, y0=ylow, y1=yhigh)

    ##### HasFace protocol #####

    def _get_brushes(self) -> list[QtGui.QBrush]:
        return self.opts["brushes"]

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        colors = []
        for brush in self._get_brushes():
            colors.append(brush.color().getRgbF())
        return np.array(colors, dtype=np.float32)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.opts["x0"]))
        brushes = self._get_brushes()
        for brush, c in zip(brushes, color):
            brush.setColor(array_to_qcolor(c))
        self.setOpts(brushes=brushes)

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_qt_brush_style(brush.style()) for brush in self._get_brushes()]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        brushes = self._get_brushes()
        if isinstance(pattern, Hatch):
            ptn = to_qt_brush_style(pattern)
            for brush in brushes:
                brush.setStyle(ptn)
        else:
            for brush, ptn in zip(brushes, pattern):
                brush.setStyle(to_qt_brush_style(ptn))
        self.setOpts(brushes=brushes)

    ##### HasEdges protocol #####

    def _get_pens(self) -> list[QtGui.QPen]:
        return self.opts["pens"]

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        colors = []
        for pen in self._get_pens():
            colors.append(pen.color().getRgbF())
        return np.array(colors, dtype=np.float32)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, len(self.opts["x0"]))
        pens = self._get_pens()
        for pen, c in zip(pens, color):
            pen.setColor(array_to_qcolor(c))
        self.setOpts(pens=pens)

    def _plt_get_edge_width(self) -> float:
        return np.array([pen.widthF() for pen in self._get_pens()])

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if np.isscalar(width):
            width = np.full(len(self.opts["x0"]), width)
        pens = self._get_pens()
        for pen, w in zip(pens, width):
            pen.setWidthF(w)
        self.setOpts(pens=pens)

    def _plt_get_edge_style(self) -> LineStyle:
        return [from_qt_line_style(pen.style()) for pen in self._get_pens()]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        pens = self._get_pens()
        if isinstance(style, LineStyle):
            style = to_qt_line_style(style)
            for pen in pens:
                pen.setStyle(style)
        else:
            for pen, s in zip(pens, style):
                pen.setStyle(to_qt_line_style(s))
        self.setOpts(pens=pens)

    def _plt_set_hover_text(self, texts: list[str]):
        self._hover_texts = texts

    def _plt_connect_pick_event(self, callback):
        def cb(ev: QtGui.QMouseEvent):
            idx = self.barsUnderCursor(ev.pos())
            if idx:
                callback(idx)

        self.clicked.connect(cb)

    def barsUnderCursor(self, pos: QtCore.QPointF) -> list[int]:
        rect: QtCore.QRectF = self.boundingRect()
        if not rect.contains(pos):
            return -1
        x0, x1, y0, y1 = self._plt_get_data()
        px, py = pos.x(), pos.y()
        indices = np.where((x0 <= px) & (px <= x1) & (y0 <= py) & (py <= y1))[0]
        return indices.tolist()

    def hoverEvent(self, ev: pgHoverEvent):
        vb = self.getViewBox()
        if vb is not None and self._hover_texts is not None:
            idx = self.barsUnderCursor(ev.pos())
            if idx:
                self._toolTipCleared = False
                vb.setToolTip(self._hover_texts[idx[-1]])
            else:
                if not self._toolTipCleared:
                    vb.setToolTip("")
                    self._toolTipCleared = True

    def mousePressEvent(self, ev: QtGui.QMouseEvent):
        super().mousePressEvent(ev)
        self.clicked.emit(ev)



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend.bokeh._base import (
    BokehLayer,
    from_bokeh_hatch,
    from_bokeh_line_style,
    to_bokeh_hatch,
    to_bokeh_line_style,
)
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, Hatch, LineStyle
from whitecanvas.utils.normalize import arr_color, as_color_array, hex_color
from whitecanvas.utils.type_check import is_real_number

# column names
TEXT = "text"
TEXT_SIZE = "text_font_size"
TEXT_COLOR = "text_color"
TEXT_ANGLE = "angle"
BG_COLOR = "background_fill_color"
BG_HATCH = "background_hatch_pattern"
BD_COLOR = "border_line_color"
BD_WIDTH = "border_line_width"
BD_STYLE = "border_line_dash"

INVISIBLE = "#00000000"


@check_protocol(TextProtocol)
class Texts(BokehLayer[bk_models.Text]):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        ntexts = len(text)
        self._data = bk_models.ColumnDataSource(
            data={
                "x": x,
                "y": y,
                TEXT: text,
                TEXT_SIZE: ["12pt"] * ntexts,
                TEXT_COLOR: ["black"] * ntexts,
                TEXT_ANGLE: [0] * ntexts,
                BG_COLOR: [INVISIBLE] * ntexts,
                BG_HATCH: [""] * ntexts,
                BD_COLOR: [INVISIBLE] * ntexts,
                BD_WIDTH: [0] * ntexts,
                BD_STYLE: ["solid"] * ntexts,
            }
        )
        self._model = bk_models.Text(
            x="x",
            y="y",
            text=TEXT,
            text_font="helvetica",
            text_font_size=TEXT_SIZE,
            text_color=TEXT_COLOR,
            text_align="left",
            angle=TEXT_ANGLE,
            background_fill_color=BG_COLOR,
            background_hatch_pattern=BG_HATCH,
            border_line_color=BD_COLOR,
            border_line_width=BD_WIDTH,
            border_line_dash=BD_STYLE,
        )
        self._visible = True

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible
        if visible:
            self._model.text_color = INVISIBLE
            self._model.background_fill_color = INVISIBLE
            self._model.border_line_color = INVISIBLE
        else:
            self._model.text_color = TEXT_COLOR
            self._model.background_fill_color = BG_COLOR
            self._model.border_line_color = BD_COLOR

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        return self._data.data[TEXT]

    def _plt_set_text(self, text: list[str]):
        self._data.data[TEXT] = text

    def _plt_get_text_color(self):
        return np.stack([arr_color(c) for c in self._data.data[TEXT_COLOR]])

    def _plt_set_text_color(self, color):
        color = as_color_array(color, len(self._data.data[TEXT]))
        self._data.data[TEXT_COLOR] = [hex_color(c) for c in color]

    def _plt_get_text_size(self) -> float:
        return np.array(
            [float(s.rstrip("pt")) for s in self._data.data[TEXT_SIZE]],
            dtype=np.float32,
        )

    def _plt_set_text_size(self, size: float):
        if is_real_number(size):
            size = np.full(len(self._data.data[TEXT]), size)
        self._data.data[TEXT_SIZE] = [f"{round(s, 1)}pt" for s in size]

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        return self._data.data["x"], self._data.data["y"]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        x, y = position
        cur_data = self._data.data.copy()
        cur_data["x"], cur_data["y"] = x, y
        cur_size = len(cur_data[TEXT])
        if x.size > cur_size:
            _n = x.size - cur_size
            _concat = np.concatenate
            cur_data[TEXT] = _concat([cur_data[TEXT], [""] * _n])
            cur_data[TEXT_SIZE] = _concat([cur_data[TEXT_SIZE], ["12pt"] * _n])
            cur_data[TEXT_COLOR] = _concat([cur_data[TEXT_COLOR], ["black"] * _n])
            cur_data[TEXT_ANGLE] = _concat([cur_data[TEXT_ANGLE], [0] * _n])
            cur_data[BG_COLOR] = _concat([cur_data[BG_COLOR], [INVISIBLE] * _n])
            cur_data[BG_HATCH] = _concat([cur_data[BG_HATCH], [""] * _n])
            cur_data[BD_COLOR] = _concat([cur_data[BD_COLOR], [INVISIBLE] * _n])
            cur_data[BD_WIDTH] = _concat([cur_data[BD_WIDTH], [0] * _n])
            cur_data[BD_STYLE] = _concat([cur_data[BD_STYLE], ["solid"] * _n])
        elif x.size < cur_size:
            cur_data[TEXT] = cur_data[TEXT][: x.size]
            cur_data[TEXT_SIZE] = cur_data[TEXT_SIZE][: x.size]
            cur_data[TEXT_COLOR] = cur_data[TEXT_COLOR][: x.size]
            cur_data[TEXT_ANGLE] = cur_data[TEXT_ANGLE][: x.size]
            cur_data[BG_COLOR] = cur_data[BG_COLOR][: x.size]
            cur_data[BG_HATCH] = cur_data[BG_HATCH][: x.size]
            cur_data[BD_COLOR] = cur_data[BD_COLOR][: x.size]
            cur_data[BD_WIDTH] = cur_data[BD_WIDTH][: x.size]
            cur_data[BD_STYLE] = cur_data[BD_STYLE][: x.size]
        self._data.data = cur_data

    def _plt_get_text_anchor(self) -> Alignment:
        return Alignment(self._model.text_align)

    def _plt_set_text_anchor(self, anc: Alignment):
        self._model.text_align = anc.value

    def _plt_get_text_rotation(self) -> NDArray[np.floating]:
        return np.array(self._data.data[TEXT_ANGLE], dtype=np.float32)

    def _plt_set_text_rotation(self, rotation: float | NDArray[np.floating]):
        if is_real_number(rotation):
            rotation = np.full(len(self._data.data[TEXT]), rotation)
        self._data.data[TEXT_ANGLE] = rotation

    def _plt_get_text_fontfamily(self) -> str:
        return self._model.text_font

    def _plt_set_text_fontfamily(self, fontfamily: str):
        self._model.text_font = fontfamily

    ##### HasFaces #####

    def _plt_get_face_color(self):
        return np.stack([arr_color(c) for c in self._data.data[BG_COLOR]])

    def _plt_set_face_color(self, color):
        color = as_color_array(color, len(self._data.data[TEXT]))
        self._data.data[BG_COLOR] = [hex_color(c) for c in color]

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_bokeh_hatch(h) for h in self._data.data[BG_HATCH]]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            pattern = [pattern] * len(self._data.data[TEXT])
        self._data.data[BG_HATCH] = [to_bokeh_hatch(p) for p in pattern]

    def _plt_get_edge_color(self):
        return np.stack([arr_color(c) for c in self._data.data[BD_COLOR]])

    def _plt_set_edge_color(self, color):
        color = as_color_array(color, len(self._data.data[TEXT]))
        self._data.data[BD_COLOR] = [hex_color(c) for c in color]

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return np.array(self._data.data[BD_WIDTH], dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(len(self._data.data[TEXT]), width)
        self._data.data[BD_WIDTH] = width

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_bokeh_line_style(s) for s in self._data.data[BD_STYLE]]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * len(self._data.data[TEXT])
        self._data.data[BD_STYLE] = [to_bokeh_line_style(s) for s in style]



from whitecanvas.canvas import CanvasBase, CanvasGrid


def view(grid: CanvasGrid, app: str = "qt"):
    if app == "qt":
        return _view_qt(grid)
    elif app == "tk":
        return _view_tk(grid)
    raise ValueError(f"Appication {app} not supported")


def make_dim_slider(canvas: CanvasBase, app: str = "qt"):
    if app == "qt":
        sl = _slider_qt(canvas)
    elif app == "tk":
        sl = _slider_tk(canvas)
    else:
        raise ValueError(f"Appication {app} not supported")
    return sl


def _view_qt(grid: CanvasGrid):
    from whitecanvas.backend._window._qt import QtMainWindow

    main = QtMainWindow(grid)
    main.show()
    return main


def _view_tk(grid: CanvasGrid):
    from whitecanvas.backend._window._tk import TkMainWindow

    main = TkMainWindow(grid)
    main.mainloop()
    return main


def _slider_qt(canvas: CanvasBase):
    from whitecanvas.backend._window._qt import QtDimSliders

    return QtDimSliders.from_canvas(canvas)


def _slider_tk(canvas: CanvasGrid):
    from whitecanvas.backend._window._tk import TkDimSliders

    return TkDimSliders.from_canvas(canvas)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, Sequence, TypeVar

import numpy as np
from cmap import Color, Colormap
from numpy.typing import NDArray

from whitecanvas.canvas._palette import ColorPalette
from whitecanvas.types import Hatch, LineStyle, Symbol
from whitecanvas.utils.collections import OrderedSet
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers.tabular._df_compat import DataFrameWrapper

_V = TypeVar("_V", bound=Any)
_DF = TypeVar("_DF")


class Plan(ABC, Generic[_V]):
    @abstractmethod
    def generate(
        self,
        vals: np.ndarray,
        by_all: tuple[str, ...],
    ) -> list[_V]:
        """
        Generate values for given labels.

        Length of the returned list must be equal to vals.shape[0].
        """


class CategoricalPlan(Plan[_V]):
    def __init__(self, by: tuple[str, ...]):
        self._by = by

    @property
    def by(self) -> tuple[str, ...]:
        return self._by

    @classmethod
    @abstractmethod
    def default(cls, by: Sequence[str]) -> Self:
        """Create a default plan."""


class CyclicPlan(CategoricalPlan[_V]):
    """
    Plan of cyclic values.

    This object is used to generate values for each category.
    """

    def __init__(self, by: tuple[str, ...], values: list[_V]):
        super().__init__(by)
        self._values = values

    def __repr__(self) -> str:
        cname = type(self).__name__
        return f"{cname}(by={self._by!r}, values={self._values})"

    @property
    def values(self) -> list[_V]:
        return self._values

    @classmethod
    @abstractmethod
    def _default_values(cls) -> list[_V]:
        """Default values for the plan."""

    @classmethod
    @abstractmethod
    def _norm_value(cls, v: Any) -> _V:
        """Normalize the value."""

    @classmethod
    def default(cls) -> Self:
        return cls((), cls._default_values())

    @classmethod
    def new(cls, by, values: list[_V] | None = None):
        """Create a new plan."""
        if values is None:
            values = cls._default_values()
        return cls(tuple(by), values)

    @classmethod
    def from_const(cls, value: _V) -> Self:
        """Create a plan that always returns a constant value."""
        return cls((), values=[cls._norm_value(value)])

    def is_const(self) -> bool:
        """Return True if the plan is a constant plan."""
        return len(self.by) == 0

    def is_not_const(self) -> bool:
        """Return True if the plan is not a constant plan."""
        return not self.is_const()

    def get_const_value(self) -> _V:
        """Return the constant value if the plan is a constant plan."""
        if self.is_const():
            return self.values[0]
        raise ValueError("The plan is not a constant plan.")

    def generate(
        self,
        labels: list[tuple[Any, ...]],
        by_all: tuple[str, ...],
    ) -> list[_V]:
        # labels = [("a",), ("b",)] ... the all set of unique labels
        # by_all = ("column-0", "column-1")
        indices = [by_all.index(b) for b in self.by]
        size = len(self.values)
        out_lookup: dict[tuple[Any, ...], _V] = {}
        i = 0
        for row in labels:
            row_filt = tuple(row[i] for i in indices)
            if row_filt not in out_lookup:
                out_lookup[row_filt] = self.values[i % size]
                i += 1

        ret = [out_lookup[tuple(_r[i] for i in indices)] for _r in labels]
        return ret

    def create_key_values(
        self,
        values: DataFrameWrapper[_DF],  # the data frame
    ) -> list[tuple[tuple, _V]]:
        """Map dataframe to values of the same size."""
        if self._by:
            series = [values[k] for k in self._by]
        else:
            # constant, no key filter
            return [((), self.values[0])]
        out = []
        i = 0
        for row in OrderedSet(zip(*series)):
            val = self.values[i % len(self.values)]
            out.append((row, val))
            i += 1
        return out

    def map(
        self,
        values: DataFrameWrapper[_DF],  # the data frame
    ) -> Sequence[_V]:
        """Map dataframe to values of the same size."""
        if self._by:
            series = [values[k] for k in self._by]
        else:
            # constant, no key filter
            return [self.values[0]] * len(values)
        out = np.empty(series[0].shape, dtype=object)
        i = 0
        for row in OrderedSet(zip(*series)):
            sl: NDArray[np.bool_] = np.all(
                np.column_stack([a == b for a, b in zip(series, row)]), axis=1
            )
            ntrue = sl.sum()
            val = self.values[i % len(self.values)]
            out[sl] = np.full(ntrue, val, dtype=object)
            i += 1
        return out

    def to_entries(self, df: DataFrameWrapper[_DF]) -> list[tuple[str, _V]]:
        """Prepare legend item entries."""
        if len(df) == 0:
            return []
        values = self.map(df)
        if self.by:
            entries = [
                (", ".join(str(n) for n in key), value)
                for key, value in self.create_key_values(df)
            ]
        else:
            entries = [("", values[0])]
        return entries


class MapPlan(ABC, Generic[_V]):
    def __init__(
        self,
        on: tuple[str, ...],
        mapper: Callable[[dict[str, np.ndarray]], Sequence[_V]],
    ):
        self._on = on
        self._mapper = mapper

    def __repr__(self) -> str:
        cname = type(self).__name__
        if self._on:
            return f"{cname}(on={self._on!r}, mapper={self._mapper})"
        else:
            return f"{cname}(mapper={self._mapper!r})"

    @classmethod
    def from_map(
        cls,
        on: Sequence[str],
        mapper: Callable[[dict[str, np.ndarray]], Sequence[_V]],
    ) -> Self:
        if not isinstance(on, (tuple, list)):
            raise TypeError(f"on must be a sequence, not {type(on)}")
        return cls(tuple(on), mapper)

    @classmethod
    def from_const(cls, value: _V) -> Self:
        """Create a map plan that always returns a constant value."""
        return cls.from_map((), ConstMap(value))

    def is_const(self) -> bool:
        """Return True if the plan is a constant plan."""
        return isinstance(self._mapper, ConstMap)

    def is_not_const(self) -> bool:
        """Return True if the plan is not a constant plan."""
        return not self.is_const()

    def get_const_value(self) -> _V:
        """Return the constant value if the plan is a constant plan."""
        if isinstance(self._mapper, ConstMap):
            return self._mapper._value
        raise ValueError("The plan is not a constant plan.")

    @classmethod
    def default(cls) -> Self:
        return cls((), cls._default_mapper())

    @classmethod
    @abstractmethod
    def _default_mapper(cls) -> Callable[[dict[str, np.ndarray]], Sequence[_V]]:
        """Default mapper for the plan."""

    def map(
        self,
        values: DataFrameWrapper[_DF],  # the data frame
    ) -> Sequence[_V]:
        """Calculate the values for the input data frame."""
        if self._on:
            input_dict = {k: values[k] for k in self._on}
        else:
            # constant or default size, no key filter
            try:
                k, v = next(iter(values.iter_items()))
                input_dict = {k: v}
            except StopIteration:
                input_dict = {}
        return self._mapper(input_dict)


class ScalarMapPlan(MapPlan[float]):
    @classmethod
    def from_range(
        cls,
        on: str,
        range: tuple[float, float] | None = None,
        domain: tuple[float, float] | None = None,
    ) -> Self:
        """Add a mapper that maps a range to a value."""
        return cls.from_map((on,), RangedMap(on, range, domain))

    def get_ranged_map(self) -> RangedMap | None:
        if isinstance(self._mapper, RangedMap):
            return self._mapper
        return None


class ColorPlan(CyclicPlan[Color]):
    @classmethod
    def _default_values(cls) -> list[Color]:
        return [Color("black")]  # NOTE: will be updated by the canvas palette

    @classmethod
    def _norm_value(cls, v: Any) -> Color:
        return Color(v)

    @classmethod
    def from_palette(
        cls,
        by: Sequence[str],
        palette: ColorPalette | None = None,
        maybe_const: bool = False,
    ) -> ColorPlan:
        if palette is None:
            palette = "tab10"
        palette = ColorPalette(palette)
        colors = palette.nextn(palette.ncolors, update=False)
        if len(colors) == 1 and maybe_const:
            return cls.from_const(colors[0])
        return cls(tuple(by), colors)

    # NOTE: Color instance is detected as a sequence of 4 floats
    #       so we need to override the default mapper
    def map(
        self,
        values: DataFrameWrapper[_DF],  # the data frame
    ) -> Sequence[_V]:
        if self._by:
            series = [values[k] for k in self._by]
        else:
            # constant, no key filter
            return [self.values[0]] * len(values)
        out = np.empty((series[0].size, 4), dtype=np.float32)
        i = 0
        for row in OrderedSet(zip(*series)):
            sl: NDArray[np.bool_] = np.all(
                np.column_stack([a == b for a, b in zip(series, row)]), axis=1
            )
            ntrue = sl.sum()
            if ntrue > 0:
                val = self.values[i % len(self.values)]
                out[sl] = [val] * ntrue
                i += 1
        return out


class StylePlan(CyclicPlan[LineStyle]):
    @classmethod
    def _default_values(cls) -> list[LineStyle]:
        return [
            LineStyle.SOLID,
            LineStyle.DASH,
            LineStyle.DOT,
            LineStyle.DASH_DOT,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> LineStyle:
        return LineStyle(v)


class HatchPlan(CyclicPlan[Hatch]):
    @classmethod
    def _default_values(cls) -> list[Hatch]:
        return [
            Hatch.SOLID,
            Hatch.DIAGONAL_BACK,
            Hatch.HORIZONTAL,
            Hatch.CROSS,
            Hatch.VERTICAL,
            Hatch.DIAGONAL_CROSS,
            Hatch.DOTS,
            Hatch.DIAGONAL_FORWARD,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> Hatch:
        return Hatch(v)


class SymbolPlan(CyclicPlan[Symbol]):
    @classmethod
    def _default_values(cls) -> list[Symbol]:
        return [
            Symbol.CIRCLE,
            Symbol.SQUARE,
            Symbol.TRIANGLE_UP,
            Symbol.DIAMOND,
            Symbol.TRIANGLE_DOWN,
        ]

    @classmethod
    def _norm_value(cls, v: Any) -> Symbol:
        return Symbol(v)


class ColormapPlan(MapPlan[NDArray[np.float32]]):
    @classmethod
    def _default_mapper(cls):
        return lambda x: np.ones((x.size, 4))

    @classmethod
    def from_colormap(
        cls,
        on: str,
        cmap: Colormap,
        *,
        clim: tuple[float, float] | None = None,
    ) -> Self:
        return cls.from_map((on,), ColormapMap(on, cmap, clim))

    def get_colormap_map(self) -> ColormapMap | None:
        if isinstance(self._mapper, ColormapMap):
            return self._mapper
        return None


class WidthPlan(ScalarMapPlan):
    @classmethod
    def _default_mapper(cls):
        return lambda _: 1.0


class SizePlan(ScalarMapPlan):
    @classmethod
    def _default_mapper(cls):
        return lambda _: 12.0


class AlphaPlan(ScalarMapPlan):
    @classmethod
    def _default_mapper(cls):
        return lambda _: 1.0


class ConstMap:
    def __init__(self, value):
        self._value = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{self._value!r}>"

    def __call__(self, values: dict[str, np.ndarray]) -> Sequence[_V]:
        series = next(iter(values.values()), np.zeros(0))
        return [self._value] * series.size


class RangedMap:
    def __init__(self, on, range=None, domain=None):
        _check_min_max(range)
        _check_min_max(domain)
        self._on = on
        self._range = range
        self._domain = domain

    def __call__(self, values: dict[str, np.ndarray]) -> Sequence[_V]:
        arr = values[self._on]
        if self._domain is None:
            valid = np.isfinite(arr)
            amin, amax = arr[valid].min(), arr[valid].max()
            if amin == amax:
                if self._range is not None:
                    w0, w1 = self._range
                    return np.full(arr.shape, (w0 + w1) / 2)
                else:
                    return np.full(arr.shape, amin)
        else:
            amin, amax = self._domain
        _arr = arr.clip(amin, amax)
        _arr[np.isnan(_arr)] = amin
        if self._range is not None:
            w0, w1 = self._range
            return (_arr - amin) / (amax - amin) * (w1 - w0) + w0
        else:
            return _arr

    def create_samples(self, values: dict[str, np.ndarray]) -> list[tuple[_V, _V]]:
        arr = values[self._on]
        if self._domain is None:
            valid = np.isfinite(arr)
            amin, amax = arr[valid].min(), arr[valid].max()
        else:
            amin, amax = self._domain
        if self._range is not None:
            w0, w1 = self._range
        else:
            w0, w1 = amin, amax

        if w0 == w1:
            return [((w0 + w1) / 2, amin)]
        else:
            xs = np.linspace(0, 1, 3)
            return [((x * (w1 - w0) + w0), x * (amax - amin) + amin) for x in xs]


class ColormapMap:
    def __init__(self, on, cmap: Colormap, clim=None):
        _check_min_max(clim)
        self._on = on
        self._cmap = cmap
        self._clim = clim

    def __call__(self, values: dict[str, np.ndarray]) -> Sequence[_V]:
        arr = values[self._on]
        valid = np.isfinite(arr)
        if self._clim is None:
            amin, amax = arr[valid].min(), arr[valid].max()
        else:
            amin, amax = self._clim
        if amin == amax:
            color = self._cmap(0.5)
            return np.full((arr.size, 4), np.asarray(color))
        _arr = arr.clip(amin, amax)
        _arr[np.isnan(_arr)] = amin
        return self._cmap((_arr - amin) / (amax - amin))

    def create_samples(self, values: dict[str, np.ndarray]) -> list[tuple[Color, _V]]:
        """Sample colors for the legends."""
        arr = values[self._on]
        valid = np.isfinite(arr)
        if self._clim is None:
            amin, amax = arr[valid].min(), arr[valid].max()
        else:
            amin, amax = self._clim
        if amin == amax:
            color = self._cmap(0.5)
            return [(color, amin)]
        if len(self._cmap.color_stops) > 4:
            xs = np.linspace(0, 1, 4)
        else:
            xs = self._cmap.color_stops
        return [(self._cmap(x), amin * (1 - x) + amax * x) for x in xs]


def _check_min_max(val: tuple[float, float] | None = None):
    if val is None:
        return
    mn, mx = val
    if not (is_real_number(mn) and is_real_number(mx)):
        raise TypeError(f"Must be a tuple of two real numbers, got {val!r}.")
    if mn > mx:
        raise ValueError(f"min must be less than or equal to max, got {mn} and {mx}.")
    return



"""LayerStack represents a stack of layers for n-D visualization."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Concatenate, ParamSpec

from whitecanvas.layers import _primitive
from whitecanvas.layers._base import DataBoundLayer, LayerWrapper
from whitecanvas.types import XYData, XYTextData, XYYData

if TYPE_CHECKING:
    from whitecanvas.canvas import Canvas

_T = TypeVar("_T")
_P = ParamSpec("_P")


# TODO: Generic type wrong.
class LayerStack(LayerWrapper["DataBoundLayer[_T]"], Generic[_T]):
    def __init__(
        self,
        base_layer: DataBoundLayer[_T],
        data: Slicable[_T],
        axis_names: list[str] | None = None,
    ):
        super().__init__(base_layer)
        self._data_stack = data
        if axis_names is None:
            axis_names = [f"axis_{i}" for i in reversed(range(data.ndim))]
        else:
            axis_names = list(axis_names)
        self._axis_names = axis_names

    @property
    def axis_names(self) -> list[str]:
        """List of axis names."""
        return self._axis_names.copy()

    def _update_layer_data(self, index: dict[str, int]) -> None:
        sl = tuple(index[a] for a in self._axis_names)
        data = self._get_slice(sl)
        self._base_layer._set_layer_data(data)
        self._base_layer.events.data.emit(data)

    def _get_slice(self, index) -> _T:
        return self._base_layer._norm_layer_data(self._data_stack.slice_at(index))

    def _connect_canvas(self, canvas: Canvas):
        canvas.dims.events.indices.connect(
            self._update_layer_data, unique=True, max_args=1
        )
        return super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: Canvas):
        canvas.dims.events.indices.disconnect(self._update_layer_data)
        return super()._disconnect_canvas(canvas)

    def _with_axis_names(self, axis_names: list[str]):
        stack = self._data_stack
        if len(axis_names) != stack.ndim:
            raise ValueError(
                f"Got {stack.ndim}D dataset but {len(axis_names)} axes names "
                "were given."
            )
        self._axis_names = list(axis_names)


class YLayerStack(LayerStack[NDArray[np.number]]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], DataBoundLayer[XYData]],
        x: Any,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        """Create a layer stack from the given layer constructor and data."""
        stack = _norm_one(x)
        sl = (0,) * stack.ndim
        data = stack.slice_at(sl)
        return cls(constructor(data, *args, **kwargs), stack)


class XYLayerStack(LayerStack[XYData]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], DataBoundLayer[XYData]],
        x: Any,
        y: Any,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        """Create a layer stack from the given layer constructor and data."""
        x0 = _norm_one(x)
        y0 = _norm_one(y)
        stack = StackedArray([x0, y0])
        sl = (0,) * stack.ndim
        data = stack.slice_at(sl)
        return cls(constructor(*data, *args, **kwargs), stack)


class XYYLayerStack(LayerStack[XYYData]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], DataBoundLayer[XYYData]],
        x: Any,
        y0: Any,
        y1: Any,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        """Create a layer stack from the given layer constructor and data."""
        xdata = _norm_one(x)
        ydata0 = _norm_one(y0)
        ydata1 = _norm_one(y1)
        stack = StackedArray([xdata, ydata0, ydata1])
        sl = (0,) * stack.ndim
        data = stack.slice_at(sl)
        return cls(constructor(*data, *args, **kwargs), stack)


class TextLayerStack(LayerStack[XYTextData]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], DataBoundLayer[XYTextData]],
        x: Any,
        y: Any,
        text: Any,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        """Create a layer stack from the given layer constructor and data."""
        xdata = _norm_one(x)
        ydata = _norm_one(y)
        textdata = _norm_one(text, dtype=np.str_)
        stack = StackedArray([xdata, ydata, textdata])
        sl = (0,) * stack.ndim
        data = stack.slice_at(sl)
        return cls(constructor(*data, *args, **kwargs), stack)


class ImageLayerStack(LayerStack[NDArray[np.number]]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], _primitive.Image],
        img: Any,
        rgb: bool = False,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        """Create a layer stack for multi-dimension image."""
        if rgb:
            stack = _norm_one(img, dim=3)
        else:
            stack = _norm_one(img, dim=2)
        sl = (0,) * stack.ndim
        im = stack.slice_at(sl)
        return cls(constructor(im, *args, **kwargs), stack)


class SpansLayerStack(LayerStack[NDArray[np.number]]):
    @classmethod
    def from_layer_class(
        cls,
        constructor: Callable[Concatenate[Any, Any, _P], _primitive.Spans],
        spans: Any,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ):
        stack = _norm_one(spans, dim=2)
        sl = (0,) * stack.ndim
        spans = stack.slice_at(sl)
        return cls(constructor(spans, *args, **kwargs), stack)


def _norm_one(data, dim: int = 1, dtype=np.float32) -> Slicable[NDArray[np.number]]:
    try:
        arr = np.asarray(data, dtype=dtype)
    except ValueError:
        arr = np.asarray(data, dtype=np.object_)

    if arr.ndim == dim:
        return ConstArray(arr)
    if arr.dtype == np.object_:
        # convert all the elements to 1D array
        for sl in zip(*[range(s) for s in arr.shape]):
            arr1d = np.asarray(arr[sl], dtype=dtype)
            if arr1d.ndim != dim:
                raise ValueError(
                    f"Data at index {sl} has {arr1d.ndim} dimensions but "
                    f"{dim} is required."
                )
            arr[sl] = arr1d
        return NonuniformArray(arr)
    else:
        return GridArray(arr, dim=dim)


class Slicable(ABC, Generic[_T]):
    """A class that can generate sliced data."""

    @abstractmethod
    def slice_at(self, index: tuple[int, ...]) -> _T:
        """Get the sliced data at the given index."""

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the data."""

    @property
    def ndim(self) -> int:
        """The number of dimensions of the data."""
        return len(self.shape)


class ConstArray(Slicable[_T]):
    def __init__(self, obj: _T):
        self._obj = obj

    def slice_at(self, index: tuple[int, ...]) -> _T:
        return self._obj

    @property
    def shape(self):
        return ()


class GridArray(Slicable[NDArray[np.number]]):
    def __init__(self, obj: NDArray[np.number], dim: int = 1):
        self._obj = obj
        self._dim = dim  # the ndim of data

    def slice_at(self, index: tuple[int, ...]) -> NDArray[np.number]:
        return self._obj[index]

    @property
    def shape(self):
        return self._obj.shape[: -self._dim]


class NonuniformArray(Slicable[_T]):
    def __init__(self, obj: NDArray[np.object_]):
        self._obj = obj

    def slice_at(self, index: tuple[int, ...]) -> NDArray[np.number]:
        return self._obj[index]

    @property
    def shape(self):
        return self._obj.shape


# TODO: multidimensional stripplot etc.
class TableArray(Slicable[NDArray[np.number]]):
    def __init__(self, obj: dict[Any, dict[str, NDArray[np.number]]]):
        self._obj = obj

    def slice_at(self, index: tuple[int, ...]): ...


class StackedArray(Slicable[_T]):
    """A stacked Slicable objects."""

    def __init__(
        self,
        obj: list[Slicable[_T]],
    ):
        self._objs = obj
        self._shape = ()
        for o in obj:
            if o.ndim > 0:
                if self._shape and self._shape != o.shape:
                    raise ValueError("All the objects must have the same shape or 1D.")
                self._shape = o.shape

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._objs!r})"

    def slice_at(self, index: tuple[int, ...]) -> list[_T]:
        return [obj.slice_at(index) for obj in self._objs]

    @property
    def shape(self):
        return self._shape



import numpy as np

from whitecanvas import plot as plt


def main():
    canvas = plt.figure("plotly:qt")
    x = np.linspace(0, 4, 200)
    times = np.linspace(0, 4, 10)
    multidim = canvas.dims.in_axes("time")
    multidim.add_line(x, [np.sin(x*3 - x0) for x0 in times])
    multidim.add_text([0], [0.8], [[f"t = {t:.2f}"] for t in times], size=28)
    canvas.show(block=True)

if __name__ == "__main__":
    main()



# This example shows how to define custom hover texts for markers.

from whitecanvas import new_canvas


def main():
    canvas = new_canvas(backend="matplotlib:qt")

    canvas.add_markers(
        [0, 1, 2], [0, 1, 2], color="red"
    ).with_hover_text(["first", "second", "third"])

    canvas.add_markers(
        [0, 1, 2], [1, 2, 3], color="blue"
    ).with_hover_template(
        "x: {x:.2f}, y: {y:.2f}, index: {i},\ncustom_data: {c}",
        extra=dict(c=[123, 234, 345]),
    )

    canvas.show(block=True)

if __name__ == "__main__":
    main()



# Use the `fit` function to fit a model to data

import numpy as np
from whitecanvas import new_canvas

def main():
    canvas = new_canvas(backend="matplotlib:qt")
    x = np.linspace(-1, 1, 100)
    y = x ** 2 / 2 + 1 + np.random.normal(0, 0.1, 100)
    markers = canvas.add_markers(x, y, color="lightgray")
    fit_1 = canvas.fit(markers).linear(color="blue", width=2.5)
    fit_2 = canvas.fit(markers).polynomial(2, color="red", width=2.5)

    print(fit_1, "is linear regression.")
    print(fit_2, "is polynomial regression of degree=2.")
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Union

import numpy as np
from numpy.typing import NDArray

from whitecanvas.types import Orientation

_Hint = Union[tuple[float, float], None]


def xy_size_hint(
    x: NDArray[np.number],
    y: NDArray[np.number],
    xpad_rel: float = 0.0,
    ypad_rel: float = 0.0,
) -> tuple[_Hint, _Hint]:
    if x.size == 0:
        return None, None
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()
    xpad = (xmax - xmin) * xpad_rel
    ypad = (ymax - ymin) * ypad_rel
    x_minmax = x.min() - xpad, x.max() + xpad
    y_minmax = y.min() - ypad, y.max() + ypad
    _x_hint, _y_hint = x_minmax, y_minmax
    return _x_hint, _y_hint


def xyy_size_hint(
    x: NDArray[np.number],
    y0: NDArray[np.number],
    y1: NDArray[np.number],
    orient: Orientation,
    xpad: float = 0,  # used for bar width
    ypad_rel: float = 0.0,
) -> tuple[_Hint, _Hint]:
    if x.size == 0:
        return None, None
    x_minmax = x.min() - xpad, x.max() + xpad
    ymin, ymax = min(y0.min(), y1.min()), max(y0.max(), y1.max())
    ypad = (ymax - ymin) * ypad_rel
    y_minmax = ymin - ypad, ymax + ypad
    if orient.is_vertical:
        _x_hint, _y_hint = x_minmax, y_minmax
    else:
        _x_hint, _y_hint = y_minmax, x_minmax
    return _x_hint, _y_hint


def xyz_size_hint(
    x: NDArray[np.number],
    y: NDArray[np.number],
    z: NDArray[np.number],
    xpad_rel: float = 0.0,
    ypad_rel: float = 0.0,
    zpad_rel: float = 0.0,
) -> tuple[_Hint, _Hint, _Hint]:
    if x.size == 0:
        return None, None, None
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()
    zmin, zmax = z.min(), z.max()
    xpad = (xmax - xmin) * xpad_rel
    ypad = (ymax - ymin) * ypad_rel
    zpad = (zmax - zmin) * zpad_rel
    x_minmax = x.min() - xpad, x.max() + xpad
    y_minmax = y.min() - ypad, y.max() + ypad
    z_minmax = z.min() - zpad, z.max() + zpad
    _x_hint, _y_hint, _z_hint = x_minmax, y_minmax, z_minmax
    return _x_hint, _y_hint, _z_hint



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, TypeVar

from whitecanvas.canvas.dataframe._base import BaseCatPlotter, CatIterator
from whitecanvas.layers import tabular as _lt
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.types import ColorType, Hatch, Orientation

if TYPE_CHECKING:
    from whitecanvas.canvas._base import CanvasBase

    NStr = str | Sequence[str]

_C = TypeVar("_C", bound="CanvasBase")
_DF = TypeVar("_DF")


class StackedCatPlotter(BaseCatPlotter[_C, _DF]):
    def __init__(
        self,
        canvas: _C,
        df: _DF,
        offset: str | tuple[str, ...] | None,
        value: str | None,
        orient: Orientation,
        stackby: str | tuple[str, ...] | None = None,
        update_labels: bool = False,
        numeric: bool = False,
    ):
        super().__init__(canvas, df)
        if isinstance(stackby, str):
            stackby = (stackby,)
        elif stackby is None:
            pass  # will be determined later
        else:
            if any(not isinstance(o, str) for o in stackby):
                raise TypeError(
                    "Category column(s) must be specified by a string or a sequence "
                    f"of strings, got {stackby!r}."
                )
            stackby = tuple(stackby)
        self._offset: tuple[str, ...] = offset
        self._cat_iter = CatIterator(self._df, offset, numeric=numeric)
        self._value = value
        self._stackby = stackby
        self._orient = orient
        self._update_labels = update_labels

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(offset={self._offset!r}, value={self._value!r}, "
            f"orient={self._orient!r}, stackby={self._stackby!r})"
        )

    def add_bars(
        self,
        *,
        color: ColorType | NStr | None = None,
        hatch: Hatch | NStr | None = None,
        name: str | None = None,
        extent: float = 0.8,
    ):
        canvas = self._canvas()
        df = self._df
        _splitby, _dodge = _shared.norm_dodge(
            df, self._offset, color, hatch, dodge=False
        )  # fmt: skip
        _pos_map = self._cat_iter.prep_position_map(_splitby, dodge=_dodge)

        xj = _jitter.CategoricalJitter(_splitby, _pos_map)
        yj = _jitter.IdentityJitter(self._value).check(df)

        _extent = self._cat_iter.zoom_factor(_dodge) * extent
        if not self._orient.is_vertical:
            xj, yj = yj, xj
        layer = _lt.DFBars.from_table_stacked(
            df, xj, yj, self._stackby, name=name, color=color, hatch=hatch,
            extent=_extent, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(color, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        return canvas.add_layer(layer)

    def add_area(
        self,
        *,
        color: NStr | None = None,
        width: float = 2.0,
        hatch: NStr | None = None,
        style: NStr | None = None,
        name: str | None = None,
    ):
        canvas = self._canvas()
        df = self._df
        _splitby, _dodge = _shared.norm_dodge(
            df, self._offset, color, hatch, dodge=False
        )  # fmt: skip
        _pos_map = self._cat_iter.prep_position_map(_splitby, dodge=_dodge)

        xj = _jitter.CategoricalJitter(_splitby, _pos_map)
        yj = _jitter.IdentityJitter(self._value).check(df)
        layer = _lt.DFArea.from_table_stacked(
            self._df, xj, yj, self._stackby, name=name, color=color, hatch=hatch,
            style=style, width=width, orient=self._orient, backend=canvas._get_backend()
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(color, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        return canvas.add_layer(layer)

    # def add_stem(
    #     self,
    #     *,
    #     color: NStr | None = None,
    #     symbol: Symbol = Symbol.CIRCLE,
    #     size: float | None = None,
    #     width: float | None = None,
    #     style: NStr | None = None,
    #     name: str | None = None,
    # ):
    #     ...



import numpy as np

from whitecanvas import new_canvas
from whitecanvas.core import new_jointgrid
from ._utils import assert_color_array_equal, filter_warning
import pytest

def test_cat(backend: str):
    canvas = new_canvas(backend=backend)
    rng = np.random.default_rng(1642)
    df = {
        "x": rng.normal(size=30),
        "y": rng.normal(size=30),
        "label": np.repeat(["A", "B", "C"], 10),
    }
    cplt = canvas.cat(df, "x", "y")
    cplt.add_line().update_style("--")
    cplt.add_line(color="label").update_style("label").with_markers()
    cplt.add_markers()
    cplt.add_markers(color="label")
    cplt.add_markers(hatch="label")
    cplt.add_pointplot(color="label")
    cplt.add_hist2d(bins=5)
    cplt.add_hist2d(bins=(5, 4))
    cplt.add_hist2d(bins="auto")
    cplt.add_hist2d(bins=(5, 4), color="label")
    cplt.add_hist2d(bins=("auto", 5))
    cplt.add_kde2d()
    cplt.add_kde2d(color="label")
    cplt.along_x().add_hist(bins=5)
    cplt.along_x().add_hist(bins=5, color="label")
    cplt.along_y().add_hist(bins=6)
    hist = cplt.along_y().add_hist(bins=6, color="label")
    cplt.along_x().add_kde()
    kde = cplt.along_x().add_kde(color="label")
    cplt.along_x().add_rug()
    with filter_warning(backend, "plotly"):
        cplt.along_x().add_rug(color="label")
    hist.update_color("black")
    kde.update_color("label")
    hist.update_width(1.5)
    kde.update_width(1.5)
    hist.update_style(":")
    kde.update_style("label")
    hist.update_hatch("label")
    kde.update_hatch("/")
    with filter_warning(backend, "vispy"):
        canvas.add_legend()

@pytest.mark.parametrize("orient", ["v", "h"])
def test_cat_plots(backend: str, orient: str):
    canvas = new_canvas(backend=backend)
    df = {
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
        "c": ["P", "Q"] * 15,
        "val": np.cos(np.arange(30) / 10),
    }
    if orient == "v":
        cat_plt = canvas.cat_x(df, "label", "y")
    else:
        cat_plt = canvas.cat_y(df, "y", "label")
    cat_plt.add_stripplot(color="c", alpha=0.8).move(0.1)
    cat_plt.add_swarmplot(color="c", alpha="val").move(0.1)
    cat_plt.add_boxplot(color="c").with_edge().with_outliers(ratio=0.5)
    with filter_warning(backend, "plotly"):
        box = cat_plt.add_boxplot(color="c").as_edge_only().move(0.1)
        box.update_color_palette(["blue", "red"], alpha=0.9, cycle_by="c")
        box = cat_plt.add_boxplot(hatch="c").as_edge_only().move(0.1)
        box.update_hatch_palette(["/", "x"])
        box.update_const(color="black", hatch="+")
    cat_plt.add_violinplot(color="c").with_rug()
    cat_plt.add_violinplot(color="c").with_outliers(ratio=0.5)
    cat_plt.add_violinplot(color="c").with_box()
    cat_plt.add_violinplot(color="c").as_edge_only().move(0.1).with_strip()
    cat_plt.add_violinplot(color="c").with_swarm()
    cat_plt.add_pointplot(color="c").err_by_se().err_by_sd().err_by_quantile().est_by_mean().est_by_median().move(0.1)
    cat_plt.add_barplot(color="c").err_by_se().err_by_sd().err_by_quantile().est_by_mean().est_by_median().move(0.1)
    with filter_warning(backend, "plotly"):
        cat_plt.add_rugplot(color="c").scale_by_density().move(0.1)
    cat_plt.add_heatmap_hist(bins=4)
    cat_plt.add_heatmap_hist(bins=4, color="c")

def test_markers(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "x": np.arange(30),
        "y": np.arange(30),
        "size": np.arange(30) / 2 + 8,
        "label0": np.repeat(["A", "B", "C"], 10),  # [A, A, ..., B, B, ..., C, C, ...]
        "label1": ["One"] * 10 + ["Two"] * 20,
    }

    _c = canvas.cat(df, "x", "y")
    out = _c.add_markers(color="label0", size="size", symbol="label1")
    assert len(set(out._base_layer.symbol[:10])) == 1
    assert len(set(out._base_layer.symbol[10:])) == 1

    out = _c.add_markers(color="label1", size="size", hatch="label0")
    assert len(set(out._base_layer.face.hatch[:10])) == 1
    assert len(set(out._base_layer.face.hatch[10:20])) == 1
    assert len(set(out._base_layer.face.hatch[20:])) == 1

    out = _c.add_markers(color="label1").with_edge(color="label0")
    assert len(np.unique(out._base_layer.edge.color[:10], axis=0)) == 1
    assert len(np.unique(out._base_layer.edge.color[10:20], axis=0)) == 1
    assert len(np.unique(out._base_layer.edge.color[20:], axis=0)) == 1

    # test scalar color
    out = _c.add_markers(color="black")
    assert_color_array_equal(out._base_layer.face.color, "black")

    out = _c.add_markers(color="transparent").update_edge_colormap("size")
    _c.mean_for_each("label0").add_markers(symbol="D")

def test_cat_xy(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "x": ["A", "B", "A", "B", "A", "B"],
        "y": ["P", "P", "Q", "Q", "R", "R"],
        "z": [1, 2, 3, 4, 5, 6],
    }
    im = canvas.cat_xy(df, "x", "y").first().add_heatmap(value="z")
    im.cmap
    im.cmap = "jet"
    im.clim
    im.clim = (0, 1)
    im.with_text()
    im.with_colorbar(orient="horizontal")

    df = {
        "x": ["A", "A", "A", "B", "A", "B"],
        "y": ["P", "Q", "Q", "Q", "P", "Q"],
        "z": [1.1, 2.1, 3.4, 6.4, 1.1, 6.8],
    }
    im = (
        canvas.cat_xy(df, "x", "y")
        .mean()
        .add_heatmap(value="z", fill=-1)
    )
    im.with_text(fmt=".1f")
    assert im.clim == (1.1, 6.6)

    canvas.cat_xy(df, "x", "y").first().add_markers(value="z")

    if backend != "vispy":
        canvas.add_legend()


@pytest.mark.parametrize("orient", ["v", "h"])
def test_agg(backend: str, orient: str):
    canvas = new_canvas(backend=backend)
    df = {
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
        "c": ["P", "Q"] * 15,
    }
    if orient == "v":
        cat_plt = canvas.cat_x(df, "label", "y")
    else:
        cat_plt = canvas.cat_y(df, "y", "label")
    cat_plt.mean().add_line(color="c", alpha=0.8)
    cat_plt.mean().add_markers(color="c", alpha=0.8)
    cat_plt.mean().add_bars(color="c", alpha=0.8).as_edge_only()
    cat_plt.std().add_line(color="c", width=2.5)
    cat_plt.sum().add_line()
    cat_plt.median().add_bars(color="c").update_width(1.5)
    cat_plt.max().add_markers()
    cat_plt.min().add_bars(color="c").update_hatch("/")
    cat_plt.first().add_line(color="c")
    if orient == "v":
        canvas.cat_x(df, x="label").count().add_line(color="c")
    else:
        canvas.cat_y(df, y="label").count().add_line(color="c")

    cat_plt.mean_for_each("c").add_stripplot()
    cat_plt.median_for_each("c").add_stripplot()
    cat_plt.std_for_each("c").add_stripplot()
    cat_plt.sum_for_each("c").add_stripplot()
    cat_plt.min_for_each("c").add_stripplot()
    cat_plt.max_for_each("c").add_stripplot()

def test_cat_legend(backend: str):
    if backend == "vispy":
        pytest.skip("vispy does not support legend")
    canvas = new_canvas(backend=backend)
    df = {
        "x": np.arange(30),
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
    }

    _c = canvas.cat(df, "x", "y")
    _c.add_line(color="label").move(0.1, 0.1)
    _c.add_markers(color="label").move(0.1, 0.1)
    canvas.add_legend()

def test_catx_legend(backend: str):
    if backend == "vispy":
        pytest.skip("vispy does not support legend")
    canvas = new_canvas(backend=backend)
    df = {
        "x": ["P", "Q"] * 15,
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
    }
    _c = canvas.cat_x(df, "x", "y")
    _c.add_boxplot(color="label", alpha=0.8)
    _c.add_violinplot(color="label", alpha=0.8).with_rug()
    _c.add_pointplot(color="label", alpha=0.8).err_by_se()
    _c.add_barplot(color="label", alpha=0.8)
    canvas.add_legend()

def test_marker_legend():
    canvas = new_canvas("mock")
    df = {
        "x": ["P", "Q"] * 15,
        "y": np.arange(30),
        "z": np.sin(np.arange(30) / 10),
        "label": np.repeat(["A", "B", "C"], 10),
    }
    canvas.cat_x(df, "x", "y").add_stripplot(color="label")
    canvas.cat_x(df, "x", "y").add_stripplot().update_size("z")
    canvas.cat_x(df, "x", "y").add_stripplot().update_colormap("z")
    canvas.cat_x(df, "x", "y").add_stripplot(symbol="label")
    canvas.cat_x(df, "x", "y").add_stripplot(hatch="label")
    canvas.add_legend()

def test_single_value():
    canvas = new_canvas("mock")
    df = {
        "x": ["P", "Q"] * 15,
        "y": np.arange(30),
        "z": np.ones(30),
        "label": np.repeat(["A", "B", "C"], 10),
    }
    canvas.cat_x(df, "x", "y").add_stripplot(color="label")
    canvas.cat_x(df, "x", "y").add_stripplot().update_size("z")
    canvas.cat_x(df, "x", "y").add_stripplot().update_colormap("z")
    canvas.add_legend()

@pytest.mark.parametrize("orient", ["v", "h"])
def test_numeric_axis(backend: str, orient: str):
    canvas = new_canvas(backend=backend)
    df = {
        "y": np.arange(30),
        "label": np.repeat([2, 5, 6], 10),
        "c": ["P", "Q"] * 15,
    }
    if orient == "v":
        cat_plt = canvas.cat_x(df, "label", "y", numeric_axis=True)
    else:
        cat_plt = canvas.cat_y(df, "y", "label", numeric_axis=True)
    cat_plt.add_stripplot(color="c", dodge=True).move(0.1)
    cat_plt.add_swarmplot(color="c").move(0.1)
    cat_plt.add_boxplot(color="c").move(0.1).with_outliers(ratio=0.5)
    with filter_warning(backend, "plotly"):
        cat_plt.add_boxplot(color="c").as_edge_only()
    cat_plt.add_violinplot(color="c").move(0.1).with_rug()
    cat_plt.add_violinplot(color="c").with_outliers(ratio=0.5)
    cat_plt.add_violinplot(color="c").with_box()
    cat_plt.add_violinplot(color="c").as_edge_only().with_strip()
    cat_plt.add_violinplot(color="c").with_swarm()
    cat_plt.add_pointplot(color="c").move(0.1).err_by_se().err_by_sd().err_by_quantile().est_by_mean().est_by_median()
    cat_plt.add_barplot(color="c").move(0.1).err_by_se().err_by_sd().err_by_quantile().est_by_mean().est_by_median()
    with filter_warning(backend, "plotly"):
        cat_plt.add_rugplot(color="c").move(0.1).scale_by_density()

def test_stack(backend: str):
    canvas = new_canvas(backend=backend)
    df = {
        "y": np.arange(30),
        "label": np.repeat([2, 5, 6], 10),
        "c": ["P", "Q"] * 15,
    }
    cat_plt = canvas.cat_x(df, "label", "y", numeric_axis=True)
    cat_plt.stack("c").add_bars(color="c")
    area = cat_plt.stack("c").add_area(hatch="c")
    area.update_color("c")
    area.update_color("black")
    area.update_hatch("/")
    area.update_style("--")
    area.update_style("c")
    area.move(0.1, 0.1)
    with filter_warning(backend, "vispy"):
        canvas.add_legend()

def test_joint_cat(backend: str):
    joint = new_jointgrid(backend=backend, loc=(0, 0), size=(180, 180))
    df = {
        "x": np.arange(30),
        "y": np.arange(30),
        "c": np.repeat(["A", "B", "C"], 10),
    }
    joint.cat(df, "x", "y").add_hist2d()
    joint.cat(df, "x", "y").add_markers(color="c")

def test_pandas_and_polars():
    import pandas as pd
    import polars as pl

    canvas = new_canvas(backend="mock")
    _dict = {
        "y": np.arange(30),
        "label": np.repeat(["A", "B", "C"], 10),
        "c": ["P", "Q"] * 15,
    }
    df_pd = pd.DataFrame(_dict)
    df_pl = pl.DataFrame(_dict)

    cat_pd = canvas.cat_x(df_pd, "label", "y")
    cat_pl = canvas.cat_x(df_pl, "label", "y")
    cat_pd.add_swarmplot(color="c")
    cat_pd.mean().add_markers(color="c")
    cat_pd.first().add_markers(color="c")

    cat_pl.add_swarmplot(color="c")
    cat_pl.sort().add_stripplot(color="c")
    cat_pl.mean().add_markers(color="c")
    cat_pl.first().add_markers(color="c")



import numpy as np

from whitecanvas import new_canvas


def main():
    canvas = new_canvas("matplotlib:qt")
    rng = np.random.default_rng(14632)

    canvas.add_markers(
        rng.random(30) * 10,
        rng.random(30) * 10,
    ).with_face_multi(
        color=rng.random((30, 3)),
    ).with_edge_multi(
        color=rng.random((30, 3)),
        width=rng.random(30) * 2 + 1.0,
    ).with_size_multi(
        rng.random(30) * 30 + 5,
    )

    canvas.show(block=True)

if __name__ == "__main__":
    main()
